from typing import Optional, Mapping, Sequence, Any

from ..t import In, Out
from ..circuit import Circuit
from ..wire import wire


def build_output_port_names(
        annotations: Mapping[str, Any],
        output_port_names: Optional[Sequence[str]] = None,
    ) -> Sequence[str]:

    try:
        r_type = annotations['return']
    except KeyError:
        raise ValueError('No return annotation') from None

    if output_port_names is None:
        output_port_names = []
        if isinstance(r_type, tuple):
            for i, _ in enumerate(r_type):
                output_port_names.append(f'O{i}')
        else:
            output_port_names.append('O')
    elif isinstance(r_type, tuple):
        if len(r_type) != len(output_port_names):
            raise ValueError(f'recieved {len(output_port_names)} names expected {len(annotation)}')
    elif len(output_port_names) != 1:
        raise ValueError(f'recieved {len(output_port_names)} names expected 1')

    return output_port_names


def build_io_args(
        annotations: Mapping[str, Any],
        output_port_names: Optional[Sequence[str]] = None,
        ) -> Mapping[str, Any]:
    io_args = {}

    output_port_names = build_output_port_names(annotations, output_port_names)

    for param, annotation in annotations.items():
        if param == "return":
            if isinstance(annotation, tuple):
                for name, elem in zip(output_port_names, annotation):
                    if not isinstance(name, str):
                        raise TypeError(f'port names must be strings recieved {name}::{type(name)}')
                    io_args[name] = Out(elem)
            else:
                io_args[output_port_names[0]] = Out(annotation)

            continue
        annotation = In(annotation)
        io_args[param] = annotation
    return io_args


def build_call_args(io, annotations):
    return [getattr(io, param) for param in annotations if param != "return"]


def wire_call_result(io, call_result, annotations, output_port_names=None):
    output_port_names = build_output_port_names(annotations, output_port_names)

    if isinstance(call_result, Circuit):
        if not len(call_result.interface.outputs()) == 1:
            raise TypeError(
                "Expected register return instance with one output")
        call_result = call_result.interface.outputs()[0]

    if isinstance(annotations["return"], tuple):
        for i, name in enumerate(output_port_names):
            wire(getattr(io, name), call_result[i])
    else:
        wire(getattr(io, output_port_names[0]), call_result)



import abc
import functools
import logging as py_logging

from magma.config import config, EnvConfig
from magma.debug import debug_wire, debug_unwire
from magma.logging import root_logger, StagedLogRecord
from magma.when import (
    get_curr_block as get_curr_when_block,
    no_when,
    temp_when
)


class TraceRecursionError(RecursionError):
    def __init__(self, value):
        msg = (
            f"RecursionError when calling trace on {value}, do you have a "
            "combinational loop?"
        )
        super().__init__(msg)


config._register(
    rewire_log_level=EnvConfig("MAGMA_REWIRE_LOG_LEVEL", "WARNING")
)


_logger = root_logger()


class WiringLogRecord(StagedLogRecord):
    def __init__(self, tpl: str, *values):
        super().__init__(tpl)
        self._values = values

    @staticmethod
    def _get_debug_name(value):
        if isinstance(value, int):
            return value
        return value.debug_name

    def args(self):
        return [
            WiringLogRecord._get_debug_name(value) for value in self._values
        ]


# TODO(rsetaluri): Remove this alias.
WiringLog = WiringLogRecord


class Wire:
    """
    Wire implements wiring.

    Each wire is represented by a bit.
    """

    def __init__(self, bit):
        self._bit = bit
        self._driving = []
        self._driver = None

    def __repr__(self):
        return repr(self._bit)

    def __str__(self):
        return str(self._bit)

    def anon(self):
        return self._bit.anon()

    def unwire(self, other=None, debug_info=None):
        if other is None:
            if self._driver is None:
                _logger.warning(
                    WiringLog("Unwire called on undriven value {}, ignoring",
                              self._bit),
                    debug_info=debug_info
                )
                return
            other = self._driver
        other._driving.remove(self)
        self._driver = None

    def connect(self, other, debug_info):
        """
        Connect two wires, self should be an input and other should be an
        output, or both should be inouts
        """
        if self._driver is not None:
            # TODO(rsetaluri): Issue a different warning in the case of the old
            # driver being a "conditional" driver (i.e. from a when context).
            old_driver = self._driver
            self.unwire(old_driver)
            _logger.log(
                py_logging.getLevelName(config.rewire_log_level),
                WiringLog(
                    ("Wiring multiple outputs to same wire, using last "
                     "connection. Input: {}, Old Output: {}, New Output: {}"),
                    self._bit, old_driver._bit, other._bit),
                debug_info=debug_info
            )
        if self._bit.is_output():
            _logger.error(
                WiringLog("Using `{}` (an output) as an input", self._bit),
                debug_info=debug_info
            )
            return
        if other._bit.is_input():
            _logger.error(
                WiringLog("Using `{}` (an input) as an output", other._bit),
                debug_info=debug_info
            )
            return
        if self._bit.is_inout() and not other._bit.is_inout():
            _logger.error(
                WiringLog("Using `{}` (not inout) as an inout", other._bit),
                debug_info=debug_info
            )
            return
        if not self._bit.is_inout() and other._bit.is_inout():
            _logger.error(
                WiringLog("Using `{}` (not inout) as an inout", self._bit),
                debug_info=debug_info
            )
            return

        self._driver = other
        other._driving.append(self)

    def trace(self, skip_self=True):
        """
        If a value is an input or an intermediate (undirected), trace it until
        there is an input or inout (this is the source)

        Upon the first invocation (from a user), we skip the current bit (so
        we don't trace to ourselves)
        """
        if self._driver is not None:
            return self._driver.trace(skip_self=False)
        if not skip_self and (self._bit.is_output() or self._bit.is_inout()):
            return self._bit
        if (
            not skip_self and self._bit.has_children() and self._bit.driven()
        ):
            # Driven non-bulk, so trace children
            return self._bit.trace(skip_self=False)
        return None

    def value(self):
        """
        Return the bit connected to this bit. Specifically, return the bit this
        bit is driving if it exists. Else if, there is a unique "drivee" bit,
        return that bit. Otherwise, return None.
        """
        if self._driver is not None:
            return self._driver._bit
        if len(self._driving) == 1:
            return self._driving[0]._bit
        return None

    def driven(self):
        return self._driver is not None

    def wired(self):
        return self._driver or self._driving

    def driving(self):
        """
        Return a (possibly empty) list of all bits this bit is driving.
        """
        return [driving._bit for driving in self._driving]

    @property
    def driver(self):
        return self._driver

    @property
    def bit(self):
        return self._bit


def lazy_resolve_wireable_method(fn, process_result=None):
    """If parent is not resolved, invoke on parent, else invoke on ._wire"""

    @functools.wraps(fn)
    def wrapper(self, *args, **kwargs):
        if self._parent and not self._parent._resolved:
            result = getattr(self._parent, fn.__name__)(*args, **kwargs)
            if result is not None and process_result is not None:
                result = process_result(self, result)
            return result
        return getattr(self._wire, fn.__name__)(*args, **kwargs)

    return wrapper


def _get_index(self, result):
    return result[self.name.index]


class Wireable:
    def __init__(self):
        """
        * _enclosing_when_context: The when context active when this object was
                                   created.  Lazily created child objects
                                   should inherit their parent's
                                   enclosing when context using the
                                   `set_enclosing_when_context` API.
        * _wired_when_contexts: A list of when contexts in which this object
                                was conditionally wired.  Used to track the
                                contexts that should be updated when unwiring.
        """
        self._wire = Wire(self)
        self._enclosing_when_context = get_curr_when_block()
        self._wired_when_contexts = []
        self._parent = None

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    def set_enclosing_when_context(self, ctx):
        self._enclosing_when_context = ctx

    @property
    def enclosing_when_context(self):
        return self._enclosing_when_context

    @lazy_resolve_wireable_method
    def wired(self):
        pass

    # return the input or output Bit connected to this Bit
    def trace(self, skip_self=True):
        # We don't use the decorator here since we want the special exception
        # handler.
        if self._parent and not self._parent._resolved:
            result = self._parent.trace(skip_self)
            if result is not None:
                return result[self.name.index]
            return None
        try:
            return self._wire.trace(skip_self)
        except RecursionError:
            raise TraceRecursionError(self)

    # return the output Bit connected to this input Bit
    @functools.partial(lazy_resolve_wireable_method, process_result=_get_index)
    def value(self):
        pass

    @lazy_resolve_wireable_method
    def driven(self):
        pass

    @functools.partial(
        lazy_resolve_wireable_method,
        process_result=lambda self, result: [x[self.name.index] for x in result]
    )
    def driving(self):
        pass

    def _remove_from_wired_when_contexts(self):
        for ctx in self._wired_when_contexts:
            ctx.remove_conditional_wire(self)
        self._wired_when_contexts = []

    def _check_resolve_parent(self):
        """Parent must be resolved before wire/unwire"""
        if self._parent and not self._parent._resolved:
            self._parent._resolve_bulk_wire()

    @debug_unwire
    def unwire(self, o=None, debug_info=None, keep_wired_when_contexts=False):
        self._check_resolve_parent()
        if o is not None:
            o = o._wire
        self._wire.unwire(o, debug_info)
        if not keep_wired_when_contexts:
            self._remove_from_wired_when_contexts()

    def _wire_impl(self, o, debug_info):
        self._wire.connect(o._wire, debug_info)
        self.debug_info = debug_info
        o.debug_info = debug_info

    def wire(self, o, debug_info):
        self._check_resolve_parent()
        o._check_resolve_parent()
        curr_when_block = get_curr_when_block()
        is_conditional = (
            curr_when_block is not None
            and curr_when_block is not self._enclosing_when_context
        )
        if not is_conditional:
            self._wire_impl(o, debug_info)
            return
        curr_when_block.add_conditional_wire(self, o)
        self._wired_when_contexts.append(curr_when_block)

    @debug_wire
    def rewire(self, o, debug_info=None):
        self.unwire(debug_info=debug_info)
        self.wire(o, debug_info=debug_info)


class AggregateWireable(Wireable):
    def __init__(self):
        super().__init__()
        self._resolved = False

    @abc.abstractmethod
    def has_elaborated_children(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def _enumerate_children(self):
        raise NotImplementedError()

    def _get_conditional_drivee_info(self):
        """
        * wired_when_contexts: list of contexts in which this value appears as
                               conditionally driven.

        * conditional_wires: for each ctx in `wired_when_contexts`, contains a
                             list of conditional wire objects where `self` is a
                             drivee.

        * default_drivers: for each ctx in `wired_when_contexts`, contains the
                           default driver (if it exists) for `self`.
        """
        conditional_wires = []
        default_drivers = []
        for ctx in self._wired_when_contexts:
            conditional_wires.append(ctx.get_conditional_wires_for_drivee(self))
            default = ctx.root.get_default_drivers_dict().pop(self, None)
            if default:
                ctx.root.builder.remove_default_driver(self)
            default_drivers.append(default)
        return self._wired_when_contexts, conditional_wires, default_drivers

    def _rewire_conditional_children(self, wired_when_contexts,
                                     conditional_wires, default_drivers):
        for i, child in self._enumerate_children():
            for ctx, wires, default_driver in zip(wired_when_contexts,
                                                  conditional_wires,
                                                  default_drivers):
                if default_driver:
                    # Default driver is wired outside of a when context.
                    with no_when():
                        child.wire(default_driver[i])

                # Use original wired when context.
                with temp_when(ctx):
                    for wire in wires:
                        child.wire(wire.driver[i])

    def _resolve_conditional_children(self, value):
        # Save conditional wire info before unwire happens
        info = self._get_conditional_drivee_info()

        AggregateWireable.unwire(self, value)

        self._rewire_conditional_children(*info)

    def _resolve_driven_bulk_wire(self):
        # Remove bulk wire since children will now track the wiring
        value = self._wire.value()

        if self._wired_when_contexts:
            return self._resolve_conditional_children(value)

        # Else, was not wired in a when context, so children should not be
        # wired in a when context
        with no_when():
            AggregateWireable.unwire(self, value)

            for i, child in self._enumerate_children():
                child.wire(value[i])

    def _resolve_bulk_wire(self):
        """
        If a child reference is made, we "expand" a bulk wire into the
        constiuent children to maintain consistency
        """
        if self._parent and not self._parent._resolved:
            self._parent._resolve_bulk_wire()
        if self._resolved:
            return
        # Trace as far back as possible to know where to start
        source = self._wire.trace()
        if source is None:
            source = self
        source._resolved = True

        # Process values one at a time so that we avoid recursion
        # error for large wire chains
        to_process = source._wire.driving()
        while to_process:
            next = to_process.pop()
            next._resolved = True
            next._resolve_driven_bulk_wire()
            to_process.extend(next._wire.driving())

    def _should_wire_children(self, o):
        return (
            self._resolved or
            o._resolved or
            self.is_mixed() or
            o.is_mixed()
        )


def aggregate_wireable_method(fn):

    @functools.wraps(fn)
    def wrapper(self, *args, **kwargs):
        if self._resolved:
            return fn(self, *args, **kwargs)
        wireable_fn = getattr(AggregateWireable, fn.__name__)
        return wireable_fn(self, *args, **kwargs)

    return wrapper



import dataclasses
import networkx as nx
from typing import Any, Dict, Iterable, List, Optional, Set

from magma.bits import Bits
from magma.bitutils import clog2
from magma.common import MroVisitor, Stack
from magma.ref import TempNamedRef
from magma.smart.smart_bits import (
    SmartExprMeta,
    SmartExpr,
    SmartBitsExpr,
    SmartBits,
    issigned,
    SmartSignedOp,
    SmartExtendOp,
)


@dataclasses.dataclass(frozen=True)
class _Node:
    expr: SmartExpr
    children: List['_Node']

    def update(self, children: List['_Node']):
        self.children[:] = children

    __hash__ = object.__hash__


def _make_tree(expr: SmartExpr) -> _Node:
    children = list(_make_tree(arg) for arg in expr.args)
    return _Node(expr, children)


make_tree = _make_tree


class _Visitor(MroVisitor):
    def get_class(self, node: _Node) -> SmartExprMeta:
        return node.expr.__class__

    def generic_visit(self, node: _Node, *args, **kwargs):
        return list(
            self.visit(child, *args, **kwargs) for child in node.children
        )


class _Transformer(_Visitor):
    def generic_visit(self, node: _Node, *args, **kwargs):
        children = list(
            self.visit(child, *args, **kwargs) for child in node.children
        )
        node.update(children)
        return node


def _make_label(node: _Node) -> str:
    if isinstance(node.expr, SmartBitsExpr):
        return str(node.expr)
    label = str(node.expr)
    try:
        index = label.index("(")
    except ValueError:
        return label
    return label[:index]


def _write_debug_graph(root: _Node, filename: str):
    graph = nx.MultiDiGraph()
    stack = Stack()
    stack.push(root)
    while stack:
        node = stack.pop()
        graph.add_node(id(node), label=_make_label(node))
        for child in node.children:
            graph.add_edge(id(node), id(child))
            stack.push(child)
    nx.drawing.nx_pydot.write_dot(graph, filename)


write_debug_graph = _write_debug_graph


@dataclasses.dataclass(frozen=True)
class _Context:
    lhs: Optional[SmartBits]
    root: _Node

    def max_width(self, *widths) -> int:
        max_width = max(widths)
        if self.lhs is None:
            return max_width
        return max(max_width, len(self.lhs))


@dataclasses.dataclass
class _Partition:
    context: _Context
    nodes: Set[_Node] = dataclasses.field(default_factory=set, init=False)

    def add(self, node: _Node):
        self.nodes.add(node)


class _PartitionSet:
    def __init__(self):
        self._partitions = {}  # Dict[_Context, _Partition]
        self._context_lookup = {}  # Dict[_Node, _Context]

    def add(self, context: _Context, node: _Node):
        partition = self._partitions.setdefault(context, _Partition(context))
        partition.add(node)
        self._context_lookup[node] = context

    def get_nodes(self, context: _Context) -> Set[_Node]:
        return self._partitions[context].nodes

    def get_partition(self, node: _Node) -> _Partition:
        context = self._context_lookup[node]
        return self._partitions[context]

    def partitions(self) -> Iterable[_Partition]:
        return self._partitions.values()

    def debug_strings(self) -> Iterable[str]:
        for partition in self._partitions.values():
            yield f"Context: {str(partition.context)}"
            for node in partition.nodes:
                yield f"    {str(node)}"


class _PartitionGraph(_Visitor):
    def __init__(self, partitions: _PartitionSet):
        self._partitions = partitions

    def visit(self, node: _Node, context: _Context):
        self._partitions.add(context, node)
        super().visit(node, context)

    def visit_SmartComparisonOp(self, node: _Node, context: _Context):
        super().generic_visit(node, _Context(None, node))

    def visit_SmartShiftOp(self, node: _Node, context: _Context):
        self.visit(node.children[0], context)
        self.visit(node.children[1], _Context(None, node.children[1]))

    def visit_SmartExtendOp(self, node: _Node, context: _Context):
        for child in node.children:
            self.visit(child, _Context(None, child))

    def visit_SmartConcatOp(self, node: _Node, context: _Context):
        for child in node.children:
            self.visit(child, _Context(None, child))

    def visit_SmartMuxOp(self, node: _Node, context: _Context):
        for child in node.children:
            self.visit(child, _Context(None, child))

    def visit_SmartReductionOp(self, node: _Node, context: _Context):
        super().generic_visit(node, _Context(None, node))


def _max_width(partition: _Partition) -> int:
    leaves = filter(lambda n: len(n.children) == 0, partition.nodes)
    widths = (len(leaf.expr.bits) for leaf in leaves)
    return partition.context.max_width(*widths)


class _ResultWidthDeterminator(_Visitor):
    def __init__(self, partitions: _PartitionSet, widths: Dict[_Node, int]):
        self._partitions = partitions
        self._widths = widths

    def visit_SmartBitsExpr(self, node: _Node):
        self._widths[node] = len(node.expr.bits)

    def visit_SmartNAryContextualOp(self, node: _Node):
        super().generic_visit(node)
        partition = self._partitions.get_partition(node)
        self._widths[node] = _max_width(partition)

    def visit_SmartComparisonOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = 1

    def visit_SmartShiftOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = self._widths[node.children[0]]

    def visit_SmartConcatOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = sum(self._widths[child] for child in node.children)

    def visit_SmartMuxOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = max(self._widths[child] for child in node.children)

    def visit_SmartReductionOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = 1

    def visit_SmartSignedOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = self._widths[node.children[0]]

    def visit_SmartExtendOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = self._widths[node.children[0]]


class _ResultSignednessDeterminator(_Visitor):
    def __init__(self, signednesses: Dict[_Node, bool]):
        self._signednesses = signednesses

    def visit_SmartNAryContextualOp(self, node: _Node):
        super().generic_visit(node)
        signedness = all(self._signednesses[child] for child in node.children)
        self._signednesses[node] = signedness

    def visit_SmartComparisonOp(self, node: _Node):
        super().generic_visit(node)
        self._signednesses[node] = False

    def visit_SmartBitsExpr(self, node: _Node):
        self._signednesses[node] = issigned(node.expr.bits)

    def visit_SmartReductionOp(self, node: _Node):
        super().generic_visit(node)
        self._signednesses[node] = False

    def visit_SmartSignedOp(self, node: _Node):
        super().generic_visit(node)
        self._signednesses[node] = node.expr.op.signed

    def visit_SmartExtendOp(self, node: _Node):
        super().generic_visit(node)
        self._signednesses[node] = self._signednesses[node.children[0]]

    def visit_SmartConcatOp(self, node: _Node):
        super().generic_visit(node)
        self._signednesses[node] = False


class _SignednessInserter(_Transformer):
    def __init__(
            self,
            widths: Dict[_Node, int],
            signednesses: Dict[_Node, bool],
    ):
        self._widths = widths
        self._signednesses = signednesses

    def _insert_signed_op(self, node: _Node, signedness: bool) -> _Node:
        expr = SmartSignedOp(signedness, None)
        new_node = _Node(expr, [node])
        self._widths[new_node] = self._widths[node]
        self._signednesses[new_node] = signedness
        return new_node

    def visit_SmartNAryContextualOp(self, node: _Node) -> _Node:
        signedness = self._signednesses[node]
        children = [
            self._insert_signed_op(self.visit(child), signedness)
            for child in node.children
        ]
        node.update(children)
        return node

    def generic_visit(self, node: _Node) -> _Node:
        children = [
            self._insert_signed_op(self.visit(child), self._signednesses[child])
            for child in node.children
        ]
        node.update(children)
        return node


class _PushDownExtensions(_Transformer):
    def __init__(self, widths, signednesses):
        self._widths = widths
        self._signednesses = signednesses

    def _maybe_extend(self, child: _Node, width: int) -> _Node:
        diff = width - self._widths[child]
        if diff <= 0:
            return child
        signedness = self._signednesses[child]
        node = _Node(SmartExtendOp(diff, signedness, None), [child])
        self._widths[node] = width
        return node

    def visit_SmartNAryContextualOp(self, node: _Node) -> _Node:
        width = self._widths[node]
        children = [
            self._maybe_extend(self.visit(child), width)
            for child in node.children
        ]
        node.update(children)
        return node

    def visit_SmartComparisonOp(self, node: _Node) -> _Node:
        width = max(self._widths[child] for child in node.children)
        children = [
            self._maybe_extend(self.visit(child), width)
            for child in node.children
        ]
        node.update(children)
        return node

    def visit_SmartMuxOp(self, node: _Node) -> _Node:
        width = self._widths[node]
        children = [
            self._maybe_extend(self.visit(child), width)
            for child in node.children[:-1]
        ]
        sel_width = clog2(len(node.children) - 1)
        children.append(self._maybe_extend(node.children[-1], sel_width))
        node.update(children)
        return node


class _ShiftOperandWidthMatcher(_PushDownExtensions):
    def visit_SmartShiftOp(self, node: _Node) -> _Node:
        lchild, rchild = node.children
        lwidth, rwidth = map(self._widths.get, node.children)
        if lwidth == rwidth:
            return node
        if lwidth > rwidth:
            rchild = self._maybe_extend(rchild, lwidth)
            node.update([lchild, rchild])
            return node
        # TODO(rsetaluri): Figure out why this actually works...
        return node


class _Evaluator(_Visitor):
    def visit_SmartBitsExpr(self, node: _Node) -> Bits:
        return node.expr.bits.typed_value()

    def visit_SmartOp(self, node: _Node) -> Bits:
        operands = self.generic_visit(node)
        return node.expr.op(*operands)

    def visit_SmartMuxOp(self, node: _Node) -> Bits:
        operands = self.generic_visit(node)
        sel = operands.pop()
        sel_width = clog2(len(node.children) - 1)
        assert len(sel) >= sel_width
        if len(sel) > sel_width:
            sel = sel[:sel_width]
        return node.expr.op(*operands, sel)


def _determine_result_widths(lhs: SmartBits, root: _Node) -> Dict[_Node, int]:
    context = _Context(lhs, root)
    partitions = _PartitionSet()
    _PartitionGraph(partitions).visit(root, context)
    widths = {}
    _ResultWidthDeterminator(partitions, widths).visit(root)
    return widths


def _determine_result_signednesses(root: _Node) -> Dict[_Node, bool]:
    signednesses = {}
    _ResultSignednessDeterminator(signednesses).visit(root)
    return signednesses


def _insert_signednesses(
        root: _Node,
        widths: Dict[_Node, int],
        signednesses: Dict[_Node, bool],
) -> _Node:
    root = _SignednessInserter(widths, signednesses).visit(root)
    return root


def _push_down_extensions(
        root: _Node,
        widths: Dict[_Node, int],
        signednesses: Dict[_Node, int],
) -> _Node:
    root = _PushDownExtensions(widths, signednesses).visit(root)
    root = _ShiftOperandWidthMatcher(widths, signednesses).visit(root)
    return root


def _evaluate(root: _Node) -> Bits:
    return _Evaluator().visit(root)


def _force_width(value: SmartBits, width: int) -> SmartBits:
    diff = len(value) - width
    if diff == 0:
        return value
    value = value.typed_value()
    if diff > 0:
        value = value[:-diff]
    else:
        value = value.ext(-diff)
    return SmartBits.from_bits(value)


def evaluate_assignment(lhs: SmartBits, rhs: SmartExpr) -> SmartBits:
    root = make_tree(rhs)
    widths = _determine_result_widths(lhs, root)
    signednesses = _determine_result_signednesses(root)
    root = _insert_signednesses(root, widths, signednesses)
    root = _push_down_extensions(root, widths, signednesses)
    evaluated = SmartBits.from_bits(_evaluate(root))
    result = _force_width(evaluated, len(lhs))
    if not rhs.name or isinstance(rhs, SmartBits):
        return result
    # SmartExpr was given a lazy name, apply it now.
    temp = type(result).undirected_t(name=rhs.name)
    temp @= result
    return temp



from magma.bits import BitsMeta, Bits
from hwtypes import BitVector


class EnumMeta(BitsMeta):
    def __new__(mcs, name, bases, attrs, **kwargs):
        # Force Bits repr for now
        attrs["orig_name"] = "Bits"
        cls = super().__new__(mcs, name, bases, attrs, **kwargs)
        fields = {}
        for key, value in attrs.items():
            if key[:2] == "__" or key == "_info_":
                continue
            if key == "orig_name":
                continue
            fields[key] = value
        if not fields:
            return cls
        for field in fields:
            if field == "N":
                # TODO: Make N unreserved
                raise ValueError("N is a reserved name in Enum")
        max_value = max(value for value in fields.values())
        num_bits = max(max_value.bit_length(), 1)
        type_ = cls[num_bits]
        type_._is_magma_enum = True
        type_.fields = fields
        for key, value in fields.items():
            setattr(type_, key, BitVector[num_bits](value))
        return type_


class Enum(Bits, metaclass=EnumMeta):
    pass



import enum
import logging as py_logging

from magma.passes.passes import DefinitionPass, pass_lambda
from magma.compile_exception import MagmaCompileException


def _has_log_above_threshold(ckt, threshold):
    levels = (level for level, _, _, _ in ckt.logs)
    return any(level >= threshold for level in levels)


class RaiseLogsAsExceptionsPass(DefinitionPass):

    class Mode(enum.Enum):
        FIRST = enum.auto()
        ALL = enum.auto()
        COLLECT = enum.auto()

    def __init__(
            self, main,
            mode: Mode = Mode.FIRST,
            threshold: int = py_logging.ERROR,
            exception_type: type = MagmaCompileException
    ):
        super().__init__(main)
        self._mode = mode
        self._threshold = threshold
        self._exception_type = exception_type
        self._ckts_with_errors = []

    def __call__(self, ckt):
        # This is an optimization so that we don't check for errors if we only
        # care to find the first circuit with errors, and we've already found
        # one such circuit.
        skip_check = (
            self._mode == RaiseLogsAsExceptionsPass.Mode.FIRST and
            self._ckts_with_errors
        )
        if skip_check or not _has_log_above_threshold(ckt, self._threshold):
            return
        self._ckts_with_errors.append(ckt)

    def done(self):
        if not self._ckts_with_errors:
            return
        if self._mode == RaiseLogsAsExceptionsPass.Mode.FIRST:
            ckt = self._ckts_with_errors[0]
            raise self._exception_type(
                f"Found circuit with errors: {ckt.name}"
            )
        if self._mode == RaiseLogsAsExceptionsPass.Mode.ALL:
            msg = (
                f"Found circuits with errors: "
                f"', '.join({ckt.name for ckt in self._ckts_with_errors})"
            )
            raise self._exception_type(msg)


raise_logs_as_exceptions_pass = pass_lambda(RaiseLogsAsExceptionsPass)



from abc import abstractmethod


__all__ = ["Compiler", "make_compiler"]


class Compiler:
    def __init__(self, main, basename, opts={}):
        self.main = main
        self.basename = basename
        self.opts = opts

    def run_pre_uniquification_passes(self):
        pass

    def compile(self):
        suffix = self.suffix()
        code = self.generate_code()
        with open(f"{self.basename}.{suffix}", "w") as f:
            f.write(code)

    @abstractmethod
    def suffix(self):
        raise NotImplementedError()

    @abstractmethod
    def generate_code(self):
        raise NotImplementedError()


def make_compiler(suffix, code_generator):
    class _NewCompiler(Compiler):
        def suffix(self):
            return suffix

        def generate_code(self):
            return code_generator(self.main)

    return _NewCompiler



from functools import lru_cache, wraps
from magma.bit import Bit
from magma.bits import UInt
from magma.circuit import declare_coreir_circuit
from magma.t import In, Out


def _declare_bfloat_circuit(name: str, ports: dict, coreir_name: str):
    return declare_coreir_circuit(name, ports, coreir_name,
                                  {"exp_bits": 8, "frac_bits": 7}, "float")


def check_len(fn):
    @wraps(fn)
    def wrapper(cls, op):
        if len(cls) != 16:
            raise NotImplementedError("Only BFloat16 supported")
        return fn(cls, op)
    return wrapper


class BFloat(UInt):
    @classmethod
    @lru_cache(maxsize=None)
    @check_len
    def _declare_unary_op(cls, op):
        return _declare_bfloat_circuit(f"magma_BFloat_{len(cls)}_{op}",
                                       {"I": In(cls), "O": Out(cls)}, op)

    @classmethod
    @lru_cache(maxsize=None)
    @check_len
    def _declare_binary_op(cls, op):
        return _declare_bfloat_circuit(f"magma_BFloat_{len(cls)}_{op}",
                                       {"I0": In(cls), "I1": In(cls),
                                        "O": Out(cls)}, op)

    @classmethod
    @lru_cache(maxsize=None)
    @check_len
    def _declare_compare_op(cls, op):
        return _declare_bfloat_circuit(f"magma_BFloat_{len(cls)}_{op}",
                                       {"I0", In(cls), "I1", In(cls),
                                        "O", Out(Bit)}, op)

    @classmethod
    @lru_cache(maxsize=None)
    @check_len
    def _declare_ite(cls, T):
        t_str = str(T)
        # Sanitize
        t_str = t_str.replace("(", "_")
        t_str = t_str.replace(")", "")
        t_str = t_str.replace("[", "_")
        t_str = t_str.replace("]", "")
        return _declare_bfloat_circuit(f"magma_BFloat_{len(cls)}_ite_{t_str}",
                                       {"I0": In(T), "I1": In(T),
                                        "S": In(Bit), "O": Out(T)}, "mux")



from magma.primitives.when import WhenBuilder
from magma.passes.passes import DefinitionPass, pass_lambda


class FinalizeWhens(DefinitionPass):
    def __call__(self, definition):
        with definition.open():
            for builder in definition._context_._builders:
                if isinstance(builder, WhenBuilder) and not builder._finalized:
                    definition._context_._placer.place(builder.finalize())


finalize_whens = pass_lambda(FinalizeWhens)



from .bit_pattern import BitPattern
from .valid import Valid
from .handshake import (ReadyValid, Producer, Consumer, Decoupled, EnqIO,
                        DeqIO, Irrevocable, Monitor, is_producer,
                        is_consumer, CreditValid)



import abc
import collections
import collections.abc
import dataclasses
from functools import wraps, partial, reduce
import operator
from typing import Any, Callable, Dict, Iterable
import warnings


class Stack:
    """Lightweight wrapper over builtin lists to provide a stack interface"""
    def __init__(self):
        self._stack = []

    def push(self, value):
        self._stack.append(value)

    def pop(self):
        return self._stack.pop()

    def peek(self):
        return self._stack[-1]

    def peek_default(self, default: Any) -> Any:
        try:
            return self._stack[-1]
        except IndexError:
            return default

    def __bool__(self) -> bool:
        return bool(self._stack)


class _Ref(object):
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value is other.value

    def __hash__(self):
        return id(self.value)


class _IdentitySetBase(collections.abc.MutableSet):
    def __init__(self, refs):
        self.refs = refs

    def __contains__(self, elem):
        return _Ref(elem) in self.refs

    def __iter__(self):
        return (ref.value for ref in self.refs)

    def __len__(self):
        return len(self.refs)

    def add(self, elem):
        self.refs.add(_Ref(elem))

    def discard(self, elem):
        raise NotImplementedError()

    def remove(self, elem):
        raise NotImplementedError()

    def pop(self):
        raise NotImplementedError()

    def clear(self):
        self.refs.clear()

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, list(self))


class IdentitySet(_IdentitySetBase):
    def __init__(self, items=()):
        refs = set(map(_Ref, items))
        super().__init__(refs)


class OrderedIdentitySet(_IdentitySetBase):
    def __init__(self, items=()):
        # NOTE(rsetaluri): We use collections.OrderedDict to mimic an ordered
        # set, to avoid implementing a custom ordered set or import one, since
        # it is not natively supported.
        refs = map(lambda x: (x, None), map(_Ref, items))
        refs = collections.OrderedDict(refs)
        super().__init__(refs)

    def add(self, elem):
        self.refs[_Ref(elem)] = None


def deprecated(func=None, *, msg=None):
    """
    Inspired by https://pybit.es/decorator-optional-argument.html.

    Marks a function as deprecated. @msg is an optional parameter to override
    the default warning.

    Examples:

        @deprecated
        def foo(...): ...

        @deprecated('Don't use bar!')
        def bar(...): ...
    """
    if func is None:
        return partial(deprecated, msg=msg)
    if msg is None:
        msg = f"{func.__name__} is deprecated"

    @wraps(func)
    def _wrapper(*args, **kwargs):
        warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)

    return _wrapper


def setattrs(obj, dct, pred=None):
    for k, v in dct.items():
        if pred is None or pred(k, v):
            setattr(obj, k, v)


class ParamDict(dict):
    """
    Hashable dictionary for simple key: value parameters
    """
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


def is_int(value):
    try:
        int(value)
    except (TypeError, ValueError):
        return False
    return True


def _make_delegate_fn(method):

    def _fn(self, *args, **kwargs):
        return getattr(self._underlying, method)(*args, **kwargs)

    return _fn


def make_delegator_cls(base):
    methods = base.__abstractmethods__

    class _Delegator(base):
        def __init__(self, underlying, *args, **kwargs):
            self._underlying = underlying

    for method in methods:
        setattr(_Delegator, method, _make_delegate_fn(method))
    # NOTE(rsetaluri): We should be using the new abc.update_abstractmethods
    # function. See https://bugs.python.org/issue41905 and
    # https://docs.python.org/3.10/library/abc.html#abc.update_abstractmethods.
    _Delegator.__abstractmethods__ = frozenset()
    return _Delegator


class IterableException(ValueError):
    pass


class EmptyIterableException(IterableException):
    pass


class NonSingletonIterableException(IterableException):
    pass


def only(lst: Iterable):
    it = iter(lst)
    try:
        value = next(it)
    except StopIteration:
        raise EmptyIterableException() from None
    try:
        new_value = next(it)
    except StopIteration:
        return value
    else:
        elements = [value, new_value] + list(it)
        raise NonSingletonIterableException(elements)


def is_empty(lst: Iterable) -> bool:
    """Checks whether the iterable @lst has any elements. If @lst is a
    generator, then it is modified and should no longer be used.
    """
    it = iter(lst)
    try:
        next(it)
    except StopIteration:
        return True
    return False


class Finalizable(abc.ABC):
    @abc.abstractmethod
    def finalize(self):
        raise NotImplementedError()


class FinalizableDelegator(Finalizable):
    def __init__(self):
        self._children = {}

    def add_child(self, key: str, child: Finalizable):
        if key in self._children:
            raise ValueError(f"key '{key}' already present")
        self._children[key] = child

    def get_child(self, key: str) -> Finalizable:
        return self._children[key]

    def finalize(self):
        for child in self._children.values():
            child.finalize()


def lca_of_types(classes):
    # NOTE(rsetaluri): This implementation is inspired by:
    #   https://stackoverflow.com/questions/58290137/how-to-find-most-recent-common-ancestor-base-type-of-several-types-in-python
    return next(iter(reduce(
        operator.and_, (collections.Counter(cls.mro()) for cls in classes))))


def slice_opts(dct: Dict, cls: type, keep: bool = False):
    get_opt = dct.__getitem__ if keep else dct.pop
    if not dataclasses.is_dataclass(cls):
        raise TypeError("Expected dataclass")
    kwargs = {}
    for name, field in cls.__dataclass_fields__.items():
        try:
            value = get_opt(name)
        except KeyError:
            continue
        kwargs[name] = value
    return cls(**kwargs)


def filter_by_key(function: Callable[[Any], bool], dct: Dict):
    return {k: v for k, v in dct.items() if function(k)}


def wrap_with_context_manager(ctx_mgr):

    def decorator(fn):

        @wraps(fn)
        def wrapper(*args, **kwargs):
            with ctx_mgr:
                return fn(*args, **kwargs)

        return wrapper

    return decorator


def replace_all(s: str, replacement_map: Dict[str, str]) -> str:
    for old, new in replacement_map.items():
        s = s.replace(old, new)
    return s


class SimpleCounter:
    def __init__(self, init: int = 0):
        self._value = init

    def value(self) -> int:
        return self._value

    def next(self) -> int:
        value = self._value
        self._value += 1
        return value


def sort_by_value(dct: Dict[Any, Any]) -> Iterable[Any]:
    return sorted(dct.keys(), key=lambda k: dct[k])


class MroVisitor(abc.ABC):
    def get_class(self, node: Any) -> type:
        return node.__class__

    def visit(self, node: Any, *args, **kwargs):
        method = None
        for cls in self.get_class(node).__mro__:
            name = f"visit_{cls.__name__}"
            method = getattr(self, name, None)
            if method is not None:
                break
        if method is None:
            method = self.generic_visit
        return method(node, *args, **kwargs)

    @abc.abstractmethod
    def generic_visit(self, node: Any, *args, **kwargs):
        raise NotImplementedError()


def prologue(prologue_fn):

    def decorator(fn):

        @wraps(fn)
        def wrapped(*args, **kwargs):
            prologue_fn(*args, **kwargs)
            return fn(*args, **kwargs)

        return wrapped

    return decorator


def epilogue(epilogue_fn):

    def decorator(fn):

        @wraps(fn)
        def wrapped(*args, **kwargs):
            ret = fn(*args, **kwargs)
            epilogue_fn(*args, **kwargs)
            return ret

        return wrapped

    return decorator



import dataclasses
from typing import Optional


@dataclasses.dataclass(frozen=True)
class CompileToMlirOpts:
    flatten_all_tuples: bool = False
    basename: Optional[str] = None
    use_native_bind_processor: bool = False
    verilog_prefix: Optional[str] = None
    user_namespace: Optional[str] = None
    disable_initial_blocks: bool = False
    elaborate_magma_registers: bool = False
    extend_non_power_of_two_muxes: bool = False
    disallow_duplicate_symbols: bool = False
    location_info_style: str = "none"
    explicit_bitcast: bool = False
    disallow_expression_inlining_in_ports: bool = False
    disallow_local_variables: bool = False



from magma.smart.smart_bits import SmartBits
from magma.type_utils import TypeTransformer, issint
from magma.value_utils import ValueTransformer, make_selector
from magma.t import Direction


class _SmartifyTypeTransformer(TypeTransformer):
    def visit_Bits(self, T):
        signed = issint(T)
        return SmartBits[len(T), signed]


class _InitializeSmartValueTransformer(ValueTransformer):
    def __init__(self, value):
        self._value = value

    def visit_Bits(self, value):
        sel = make_selector(value)
        init = sel.select(self._value)
        signed = issint(type(value))
        return SmartBits[len(value), signed](init)


def make_smart(value):
    Tsmart = _SmartifyTypeTransformer().visit(type(value))
    Tsmart = Tsmart.qualify(Direction.Undirected)
    smart_value = Tsmart()
    return _InitializeSmartValueTransformer(value).visit(smart_value)



"""
Bit type primitives factored out into a separate file.
NOTE(leonardt): This avoids a circular dependency between circuit.py, bit.py,
and clock.py
"""
from functools import lru_cache
import operator

import hwtypes as ht

from magma.bit import Bit
from magma.debug import magma_helper_function
from magma.digital import Digital
from magma.circuit import Circuit, coreir_port_mapping
from magma.interface import IO
from magma.language_utils import primitive_to_python
from magma.t import In, Out


@classmethod
@lru_cache(maxsize=None)
def _declare_unary_op(cls, op):
    assert cls.undirected_t is cls
    assert op == "not", f"simulate not implemented for {op}"

    class _MagmaBitOp(Circuit):
        name = f"magma_Bit_{op}"
        coreir_name = op
        coreir_lib = "corebit"
        renamed_ports = coreir_port_mapping
        primitive = True
        stateful = False

        io = IO(I=In(Bit), O=Out(Bit))

        def simulate(self, value_store, state_store):
            I = ht.Bit(value_store.get_value(self.I))
            O = int(~I)
            value_store.set_value(self.O, O)

    return _MagmaBitOp


@classmethod
@lru_cache(maxsize=None)
def _declare_binary_op(cls, op):
    assert cls.undirected_t is cls
    python_op_name = primitive_to_python(op)
    python_op = getattr(operator, python_op_name)

    class _MagmaBitOp(Circuit):
        name = f"magma_Bit_{op}"
        coreir_name = op
        coreir_lib = "corebit"
        renamed_ports = coreir_port_mapping
        primitive = True
        stateful = False

        io = IO(I0=In(Bit), I1=In(Bit), O=Out(Bit))

        def simulate(self, value_store, state_store):
            I0 = ht.Bit(value_store.get_value(self.I0))
            I1 = ht.Bit(value_store.get_value(self.I1))
            O = int(python_op(I0, I1))
            value_store.set_value(self.O, O)

    return _MagmaBitOp


# NOTE(leonardt): Monkey patched functions.
Bit._declare_unary_op = _declare_unary_op
Bit._declare_binary_op = _declare_binary_op


def make_Define(_name, port, direction):
    @lru_cache(maxsize=None)
    def DefineCorebit():
        class _Primitive(Circuit):
            renamed_ports = coreir_port_mapping
            name = f"corebit_{_name}"
            coreir_name = _name
            coreir_lib = "corebit"

            def simulate(self, value_store, state_store):
                pass

            # Type must be a bit because coreir uses Bit for the primitive.
            io = IO(**{port: direction(Bit)})
            primitive = True
            stateful = False
        return _Primitive
    return DefineCorebit


DefineUndriven = make_Define("undriven", "O", Out)
DefineUnused = make_Define("term", "I", In)


@magma_helper_function
def unused(self):
    if self.is_input() or self.is_inout():
        raise TypeError("unused cannot be used with input/inout")
    if not getattr(self, "_magma_unused_", False):
        DefineUnused()().I.wire(self)
        # "Cache" unused calls so only one is produced
        self._magma_unused_ = True


@magma_helper_function
def undriven(self):
    if self.is_output() or self.is_inout():
        raise TypeError("undriven cannot be used with output/inout")
    self.wire(DefineUndriven()().O)


# NOTE(leonardt): Monkey patched functions.
Bit.unused = unused
Bit.undriven = undriven
Digital.unused = unused
Digital.undriven = undriven



import functools
import math
from typing import List, Optional, Sequence, Tuple

import hwtypes


__all__  = ['seq2int', 'int2seq']
__all__ += ['clog2', 'clog2safe']


# TODO(rsetaluri): We should deprecate this function and invoke the hwtypes
# versions directly at the call sites.
def seq2int(l: Sequence) -> int:
    return int(hwtypes.BitVector[len(l)](l))


# TODO(rsetaluri): We should deprecate this function and invoke the hwtypes
# versions directly at the call sites.
def int2seq(i: int, n: Optional[int] = None) -> List[int]:
    bit_length = i.bit_length()
    if n is None:
        n = bit_length
    elif n < bit_length:
        raise ValueError("Insufficient number of bits")
    return hwtypes.BitVector[n](i).bits()


def clog2(x: int) -> int:
    return math.ceil(math.log(x, 2))


def clog2safe(x: int) -> int:
    if x == 1:
        return 1
    return clog2(x)



import ast
import enum
import textwrap
import inspect
import itertools
from functools import wraps
import functools
import operator
from collections import namedtuple, Counter, UserDict
import os
from typing import Callable, Dict, List, Union

from . import cache_definition
from .common import deprecated, setattrs, OrderedIdentitySet
from .interface import *
from .wire import *
from .config import get_debug_mode, set_debug_mode
from .debug import get_debug_info, debug_info
from .is_definition import isdefinition
from .placer import Placer, StagedPlacer
from magma.syntax.combinational import combinational
try:
    import kratos
    from magma.syntax.verilog import combinational_to_verilog, \
        sequential_to_verilog
except ImportError:
    pass

from magma.clock import is_clock_or_nested_clock, Clock, ClockTypes
from magma.config import get_debug_mode, set_debug_mode, config, RuntimeConfig
from magma.definition_context import (
    DefinitionContext,
    definition_context_manager,
    push_definition_context,
    pop_definition_context,
    get_definition_context,
)
from magma.find_unconnected_ports import find_and_log_unconnected_ports
from magma.logging import root_logger, capture_logs
from magma.protocol_type import MagmaProtocol
from magma.ref import TempNamedRef, AnonRef
from magma.t import In, Type
from magma.view import PortView
from magma.wire_container import WiringLog, AggregateWireable


__all__ = ['AnonymousCircuitType']
__all__ += ['AnonymousCircuit']

__all__ += ['CircuitType']
__all__ += ['Circuit', 'DebugCircuit']
__all__ += ['DeclareCircuit']
__all__ += ['DefineCircuit', 'EndDefine', 'EndCircuit']
__all__ += ['DefineCircuitKind']

__all__ += ['CopyInstance']
__all__ += ['circuit_type_method']
__all__ += ['circuit_generator']
__all__ += ['CircuitBuilder', 'builder_method']
__all__ += ['register_instance_callback', 'get_instance_callback']

circuit_type_method = namedtuple('circuit_type_method', ['name', 'definition'])

_logger = root_logger()

config.register(use_namer_dict=RuntimeConfig(False))


class _SyntaxStyle(enum.Enum):
    NONE = enum.auto()
    OLD = enum.auto()
    NEW = enum.auto()


def _has_definition(cls, port=None):
    if cls.instances:
        return True
    if port is None:
        interface = getattr(cls, "interface", None)
        if not interface:
            return False
        return any(_has_definition(cls, p) for p in interface.ports.values())
    if port.is_output():
        return False
    if port.is_mixed():
        # mixed direction
        return any(_has_definition(cls, p) for p in port)
    return port.driven()


def _maybe_add_default_clock(cls):
    """
    If circuit @cls needs a default input clock (requires that both (a) @cls has
    an instance that has an undriven clock port and (b) @cls *doesn't* already
    have a default clock input), then we add one with the name "CLK". If a port
    with name "CLK" already exists then we do not add any port.
    """
    # Check if @cls already has a default clock.
    for port in cls.io.ports.values():
        if is_clock_or_nested_clock(type(port), (Clock,)):
            return

    # Check all instances in @cls for an undriven clock port.
    inst_ports = itertools.chain(
        *(instance.interface.ports.values() for instance in cls.instances
          if getattr(instance, "stateful", True)))
    for port in inst_ports:
        T = type(port)
        if is_clock_or_nested_clock(T, (Clock,)) and not port.driven():
            break  # undriven clock found
    else:  # no undriven clock found
        return
    if "CLK" in cls.io.ports.keys():
        # What should we do in the case that the CLK name is taken, but isn't a
        # clock type? Seems odd, but we shouldn't cause an error. For now, we
        # can assume that the user might be using some custom "Clock" type and
        # not introduce a default.  Alternatively we could add a different name.
        return

    cls.io.add("CLK", In(Clock))


def _get_interface_type(cls, add_default_clock):
    if hasattr(cls, "IO") and hasattr(cls, "io"):
        _logger.warning("'IO' and 'io' should not both be specified, ignoring "
                        "'io'", debug_info=cls.debug_info)
    # TODO(setaluri): Simplify this logic and remove InterfaceKind check
    # (this is a artifact of the circuit_generator decorator).
    if hasattr(cls, 'IO'):
        cls._syntax_style_ = _SyntaxStyle.OLD
        _logger.warning("'IO = [...]' syntax is deprecated, use "
                        "'io = IO(...)' syntax instead",
                        debug_info=cls.debug_info)
        if isinstance(cls.IO, InterfaceKind):
            return None
        return make_interface(*cls.IO)
    if hasattr(cls, "io"):
        cls._syntax_style_ = _SyntaxStyle.NEW
        # Assume circuit is stateful, but if it's set stateful=False, then we
        # don't need to add clock (useful for performance to avoid the
        # traversal)
        if add_default_clock and getattr(cls, "stateful", True):
            _maybe_add_default_clock(cls)
        return cls.io.make_interface()
    return None


def _setup_interface(cls):
    IO = _get_interface_type(cls, add_default_clock=True)
    if not hasattr(cls, "IO") and not hasattr(cls, "io"):
        return
    if IO is not None:
        cls.IO = IO
    cls.interface = cls.IO(defn=cls, renamed_ports=cls._renamed_ports_)
    setattrs(cls, cls.interface.ports, lambda k, v: isinstance(k, str))


def _add_intermediate_value(value):
    """
    Add an intermediate value `value` to `values`, handling members of
    recursive types.  Used by `_get_intermediate_values` as part of
    `CircuitKind.__repr__`.
    """
    root = value.name.root()
    # If it is a TempNamedRef which doesn't refer to a constant, then we seed
    # the set with the root value. Otherwise, we can proceed with the empty
    # set.
    if isinstance(root, TempNamedRef) and not root.value().const():
        return OrderedIdentitySet((root.value(),))
    return OrderedIdentitySet()


def _get_intermediate_values(value):
    """
    Retrieve the intermediate values connected to `value` and put them into
    `values`

    Used in the implementation of `CircuitKind.__repr__` to emit the temporary
    values in a circuit.
    """
    if value.is_output():
        return OrderedIdentitySet()
    if value.is_mixed():
        # Mixed
        return functools.reduce(operator.or_,
                                (_get_intermediate_values(v) for v in value),
                                OrderedIdentitySet())
    driver = value.value()
    if driver is None:
        return OrderedIdentitySet()
    if (
            isinstance(driver, AggregateWireable) and
            driver.name.anon() and
            driver.has_elaborated_children()
    ):
        return functools.reduce(
            operator.or_, (_get_intermediate_values(v) for v, _ in
                           value.connection_iter()),
            OrderedIdentitySet())
    values = OrderedIdentitySet()
    while driver is not None:
        values |= _add_intermediate_value(driver)
        if not driver.driven():
            break
        value = driver
        driver = driver.value()
    return values


class NamerDict(UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._inferred_names = {}

    def _set_name(self, key, value):
        if isinstance(value, (Type, MagmaProtocol)):
            key = TempNamedRef(key, value)
        value.name = key

    def _check_unique_name(
        self,
        key: str,
        value: Union[Type, 'Circuit'],
    ):
        """If key has been seen more than once, "uniquify" the names by append
           _{i} to them."""
        values = self._inferred_names.setdefault(key, [])
        if len(values) == 0:
            self._set_name(key, value)
        elif any(value is x for x in values):
            return  # already uniquified
        else:
            if len(values) == 1:
                # Make the first value consistent by appending _0.
                self._set_name(f"{key}_0", values[0])
            self._set_name(f"{key}_{len(values)}", value)
        values.append(value)

    def _set_value_name(self, key: str, value: Type):
        if not hasattr(value, "name"):
            return  # interface object is a Type without a name
        if isinstance(value.name, AnonRef):
            self._check_unique_name(key, value)
        elif isinstance(value.name, TempNamedRef):
            self._check_unique_name(value.name.name, value)

    def _set_inst_name(self, key: str, value: 'Circuit'):
        if not value.name:
            self._check_unique_name(key, value)
        else:
            self._check_unique_name(value.name, value)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if isinstance(value, (Type, MagmaProtocol)):
            self._set_value_name(key, value)
        elif isinstance(type(value), CircuitKind):
            # NOTE(leonardt): we check type(value) because this code is run in
            # the Circuit class creation pipeline (so Circuit may not be defined
            # yet).
            self._set_inst_name(key, value)

    def __hash__(self):
        return hash(tuple(sorted(self.items())))


class CircuitKind(type):
    def __prepare__(name, bases, **kwargs):
        ctx = DefinitionContext(StagedPlacer(name))
        push_definition_context(ctx, use_staged_logger=True)
        if config.use_namer_dict:
            return NamerDict()
        return type.__prepare__(name, bases, **kwargs)

    """Metaclass for creating circuits."""
    def __new__(metacls, name, bases, dct):
        if isinstance(dct, NamerDict):
            dct = dict(dct)  # resolve to dict, no longer need name logic
        # Override class name if supplied (and save class name).
        cls_name = dct.get("_cls_name_", name)
        name = dct.setdefault('name', name)

        dct.setdefault("_circuit_base_", False)
        dct.setdefault('renamed_ports', {})
        dct.setdefault('primitive', False)
        dct.setdefault('coreir_lib', 'global')
        dct.setdefault("inline_verilog_strs", [])
        dct.setdefault("coreir_metadata", {})
        dct["inline_verilog_generated"] = False
        dct["bind_modules"] = {}
        dct["compiled_bind_modules"] = {}

        # If in debug_mode is active and debug_info is not supplied, attach
        # callee stack info.
        dct.setdefault("debug_info", None)
        if get_debug_mode() and not dct["debug_info"]:
            callee_frame = inspect.getframeinfo(
                inspect.currentframe().f_back.f_back)
            module = inspect.getmodule(inspect.stack()[2][0])
            dct["debug_info"] = debug_info(callee_frame.filename,
                                           callee_frame.lineno, module)

        cls = type.__new__(metacls, name, bases, dct)

        for method in dct.get('circuit_type_methods', []):
            setattr(cls, method.name, method.definition)

        cls._syntax_style_ = _SyntaxStyle.NONE
        cls._renamed_ports_ = dct["renamed_ports"]
        # NOTE(rsetaluri): We first peek the definition context
        # (get_definition_context()) so that we can finalize it before popping
        # it. This is necessary because we want the unstaging of the logger to
        # happen (inside of pop) after finalization.
        try:
            context = get_definition_context()
        except IndexError:  # no staged placer
            cls._context_ = DefinitionContext(Placer(cls))
        else:
            assert context.placer.name == cls_name
            # Override staged context with '_context_' from namespace if
            # available.
            cls._context_ = dct.get("_context_", context)
            cls._context_.place_instances(cls)
            _setup_interface(cls)
            cls._context_.finalize(cls)
            pop_definition_context(use_staged_logger=True)

        return cls

    def __call__(cls, *largs, **kwargs):
        if "debug_info" not in kwargs:
            kwargs["debug_info"] = get_debug_info(3)
        self = super(CircuitKind, cls).__call__(*largs, **kwargs)
        if hasattr(cls, 'IO'):
            interface = cls.IO(inst=self, renamed_ports=cls.renamed_ports)
            self.setinterface(interface)
        return self

    def __str__(cls):
        interface = ""
        if hasattr(cls, "interface"):
            interface = ", ".join(f"{name}: {_type}"
                                  for name, _type in cls.interface.items())
            interface = f"({interface})"
        return f"{cls.__name__}{interface}"

    def __repr__(cls):
        if not hasattr(cls, 'IO'):
            return super().__repr__()

        name = cls.__name__
        args = cls.IO.args_to_str()

        if not isdefinition(cls):
            return f"{name} = DeclareCircuit(\"{name}\", {args})"
        s = f"{name} = DefineCircuit(\"{name}\", {args})\n"
        sorted_instances = sorted(cls.instances, key=lambda x: x.name)

        values = []
        for instance in sorted_instances:
            values.extend(list(instance.interface.ports.values()))
        values.extend(cls.interface.ports.values())
        intermediate_values = (_get_intermediate_values(value)
                               for value in values)
        intermediate_values = functools.reduce(operator.or_,
                                               intermediate_values,
                                               OrderedIdentitySet())
        if intermediate_values:
            intermediate_values = (f"{value.name} = {repr(value)}"
                                   for value in intermediate_values)
            s += "\n".join(intermediate_values) + "\n"
        # Emit instances.
        for instance in sorted_instances:
            s += repr(instance) + '\n'
        # Emit wires from instances.
        for instance in sorted_instances:
            s += repr(instance.interface)
        s += repr(cls.interface)
        s += "EndCircuit()"
        return s

    def rename(cls, new_name):
        if cls.verilogFile:
            raise Exception("Can not rename a verilog wrapped file")

        old_name = cls.name
        cls.name = new_name
        cls.coreir_name = new_name
        cls.verilog_name = new_name
        cls.__name__ = new_name

        # NOTE(rsetaluri): This is a very hacky way to try to rename wrapped
        # verilog. We simply replace the first instance of "module <old_name>"
        # with "module <new_name>". This ignores the possibility of "module
        # <new_name>" existing anywhere else, most likely in comments etc. The
        # more robust way to do this would to modify the AST directly and
        # generate the new verilog code.
        #
        # Instead, at the top of this function, we raise an exception for
        # verilog wrapped files. That is, for now it is considered an error to
        # try to rename a verilog wrapped circuit.
        if cls.verilogFile:
            find_str = f"module {old_name}"
            replace_str = f"module {new_name}"
            assert cls.verilogFile.find(find_str) != -1
            cls.verilogFile = cls.verilogFile.replace(find_str, replace_str, 1)

    def find(cls, defn):
        name = cls.__name__
        if not isdefinition(cls):
            return defn
        for i in cls.instances:
            type(i).find(defn)
        if name not in defn:
            defn[name] = cls
        return defn

    @deprecated(msg="cls.inline_verilog is deprecated, use m.inline_verilog"
                "instead")
    def inline_verilog(cls, inline_str, **kwargs):
        # NOTE(rsetaluri): This is a hack to avoid a circular import.
        from magma.inline_verilog import inline_verilog as m_inline_verilog
        with definition_context_manager(cls._context_):
            m_inline_verilog(inline_str, **kwargs)


class AnonymousCircuitType(object, metaclass=CircuitKind):
    """Abstract base class for circuits"""

    _circuit_base_ = True

    def __init__(self, *largs, **kwargs):
        self.kwargs = dict(**kwargs)

        if "debug_info" in self.kwargs:
            # Not an instance parameter, internal debug argument.
            del self.kwargs["debug_info"]

        if hasattr(self, 'default_kwargs'):
            for key, value in self.default_kwargs.items():
                self.kwargs.setdefault(key, value)

        self.name = kwargs['name'] if 'name' in kwargs else ""
        self.loc = kwargs['loc'] if 'loc' in kwargs else None
        if self.loc and len(self.loc) == 2:
            self.loc = (self.loc[0], self.loc[1], 0)
        self.interface = None
        self.defn = None
        self.used = False
        self.is_instance = True
        self.debug_info = kwargs.get("debug_info", None)

    def set_debug_info(self, debug_info):
        self.debug_info = debug_info

    def __str__(self):
        name = self.name
        if not name:
            name = f"AnonymousCircuitInst{id(self)}"
        if type(self) is AnonymousCircuitType:
            interface = ", ".join(
                f"{name}: {type(value)}"
                for name, value in self.interface.ports.items()
            )
            name += f"<{interface}>"
        else:
            name += f"<{type(self)}>"
        return name

    def __repr__(self):
        args = []
        for k, v in self.interface.ports.items():
            args.append(f"\"{k}\"")
            args.append(repr(v))
        args = ", ".join(args)
        typ = type(self).__name__
        if self.name:
            return f"{self.name} = {typ}({args})"
        return f"{typ}({args})"

    def __getitem__(self, key):
        return self.interface[key]

    def wireoutputs(self, outputs, debug_info):
        """Wire a list of outputs to the circuit's inputs"""
        inputs = self.interface.inputs()
        ni = len(inputs)
        no = len(outputs)
        if ni != no:
            msg = (
                f"Number of inputs is not equal to the number of outputs, "
                f"expected {ni} inputs, got {no}. Only {min(ni,no)} will be "
                f"wired.")
            _logger.warning(msg, debug_info=debug_info)
        for i in range(min(ni, no)):
            inputs[i].wire(outputs[i], debug_info)

    def wire(self, output, debug_info):
        """Wire a single output to the circuit's inputs"""
        if hasattr(output, 'interface'):
            # Wire the circuit's outputs to this circuit's inputs.
            self.wireoutputs(output.interface.outputs(), debug_info)
            return
        # Wire the output to this circuit's input (should only have 1 input).
        inputs = []
        for inp in self.interface.inputs():
            inputs.append(inp)
        ni = len(inputs)
        if ni == 0:
            msg = ("Wiring an output to a circuit with no input arguments, "
                   "skipping")
            _logger.warning(msg, debug_info=debug_info)
            return
        if ni != 1:
            _logger.warning(
                WiringLog("Wiring an output to a circuit with more than one "
                          "input argument, using the first input {}",
                          inputs[0]),
                debug_info=debug_info
            )
        inputs[0].wire(output, debug_info)

    @property
    def debug_name(self):
        defn_str = ""
        if hasattr(self, 'defn') and self.defn is not None:
            defn_str = str(self.defn.name)
        return f"{defn_str}.{self.name}"

    def __call__(input, *outputs, **kw):
        debug_info = get_debug_info(3)

        no = len(outputs)
        if len(outputs) == 1:
            input.wire(outputs[0], debug_info)
        elif len(outputs) >= 1:  # In case there are only kw
            input.wireoutputs(outputs, debug_info)

        # Wire up extra arguments, name to name.
        # TODO(rsetaluri): This code should be changed to use clock types.
        for key, value in kw.items():
            if key == 'enable':
                key = 'CE'
            elif key == 'reset':
                key = 'RESET'
            elif key == 'set':
                key = 'SET'  # NYI
            elif key == 'ce':
                key = 'CE'  # deprecated
            if hasattr(input, key):
                i = getattr(input, key)
                wire(value, getattr(input, key), debug_info)
            else:
                _logger.warning(
                    WiringLog(f"Instance {{}} does not have input {key}",
                              input),
                    debug_info=debug_info
                )

        o = input.interface.outputs()
        return o[0] if len(o) == 1 else tuple(o)

    def setinterface(self, interface):
        setattrs(self, interface.ports, lambda k, v: isinstance(k, str))
        self.interface = interface
        return self

    def on(self):
        self.used = True
        return self

    def off(self):
        self.used = False
        return self

    def rename(self, name):
        self.name = name
        return self

    def isclocked(self):
        return self.interface.isclocked()

    def clockargs(self):
        return self.interface.clockargs()

    def inputargs(self):
        return self.interface.inputargs()

    def outputargs(self):
        return self.interface.outputargs()

    @classmethod
    def open(cls):
        return definition_context_manager(cls._context_)


def AnonymousCircuit(*decl):
    """
    AnonymousCircuits are like macros - the circuit instances are not placed
    """
    if len(decl) == 1:
        decl = decl[0]
    return AnonymousCircuitType().setinterface(AnonymousInterface(decl))


class CircuitType(AnonymousCircuitType):
    _circuit_base_ = True

    """Placed circuit - instances placed in a definition"""
    def __init__(self, *largs, **kwargs):
        super(CircuitType, self).__init__(*largs, **kwargs)
        try:
            context = get_definition_context()
            context.placer.place(self)
        except IndexError:  # instances must happen inside a definition context
            raise Exception("Can not instance a circuit outside a definition")

    def __repr__(self):
        args = []
        for k, v in self.kwargs.items():
            if isinstance(v, tuple):
                # {   # Format identifier
                # 0:  # first parameter
                # #   # use "0x" prefix
                # 0   # fill with zeroes
                # {1} # to n characters (including 0x), per the second parameter
                # x   # hexadecimal number, using lowercase letters for a-f
                # }   # End of format identifier
                if len(v) == 2:
                    v = "{0:#0{1}x}".format(v[0], v[1] // 4)
            else:
                v = f"\"{v}\""
            args.append(f"{k}={v}")
        args = ", ".join(args)
        typ = type(self).__name__
        if self.name:
            return f"{self.name} = {typ}({args})"
        return f"{typ}({args})"


@deprecated(
    msg="DeclareCircuit factory method is deprecated, subclass Circuit instead")
def DeclareCircuit(name, *decl, **args):
    """DeclareCircuit Factory"""
    debug_info = get_debug_info(4)
    metacls = CircuitKind
    bases = (CircuitType,)
    dct = metacls.__prepare__(name, bases)
    dct.update(dict(IO=decl,
                    debug_info=debug_info,
                    is_definition=False,
                    primitive=args.get('primitive', True),
                    stateful=args.get('stateful', False),
                    simulate=args.get('simulate'),
                    firrtl_op=args.get('firrtl_op'),
                    circuit_type_methods=args.get('circuit_type_methods', []),
                    coreir_lib=args.get('coreir_lib', "global"),
                    coreir_name=args.get('coreir_name', name),
                    coreir_genargs=args.get('coreir_genargs', None),
                    coreir_configargs=args.get('coreir_configargs', {}),
                    verilog_name=args.get('verilog_name', name),
                    default_kwargs=args.get('default_kwargs', {}),
                    renamed_ports=args.get('renamed_ports', {})))
    return metacls(name, bases, dct)


def _get_name(name, bases, dct):
    try:
        return dct["name"]
    except KeyError:
        pass
    # Take name of the first base class which is not a base circuit type, if one
    # exists.
    for base in bases:
        if base._circuit_base_:
            continue
        if not issubclass(base, AnonymousCircuitType):
            raise Exception(f"Must subclass from AnonymousCircuitType or a "
                            f"subclass of AnonymousCircuitType ({base})")
        return base.name
    return name


class DefineCircuitKind(CircuitKind):
    def __new__(metacls, name, bases, dct):
        dct["_cls_name_"] = name  # save original name for debugging purposes
        name = _get_name(name, bases, dct.copy())
        dct["name"] = name
        dct["renamed_ports"] = dct.get("renamed_ports", {})

        self = CircuitKind.__new__(metacls, name, bases, dct)

        self.verilog = dct.get("verilog", None)
        self.verilogFile = None
        self.verilogLib = None

        self.verilog_name = dct.get('verilog_name', name)
        self.coreir_name = dct.get('coreir_name', name)
        self.coreir_genargs = dct.get('coreir_genargs', None)
        self.coreir_configargs = dct.get('coreir_configargs', {})
        self.default_kwargs = dct.get('default_kwargs', {})
        self.firrtl = None

        has_definition = _has_definition(self)
        self._is_definition = dct.get('is_definition', has_definition)
        self.is_instance = False

        run_unconnected_check = False
        if hasattr(self, "definition"):
            if self._syntax_style_ is _SyntaxStyle.OLD:
                _logger.warning("'definition' class method syntax is "
                                "deprecated, use inline definition syntax "
                                "instead", debug_info=self.debug_info)
                # NOTE(leonardt): We can call setup here before placing
                # instances because old style IO syntax doesn't support the
                # automatic clock lifting anyways
                _setup_interface(self)
                with definition_context_manager(self._context_):
                    self.definition()
                self._context_.place_instances(self)
                self._context_.finalize(self)
                self._is_definition = True
                run_unconnected_check = True
            elif self._syntax_style_ is _SyntaxStyle.NEW:
                _logger.warning("Supplying method 'definition' with new inline "
                                "definition syntax is not supported, ignoring "
                                "'definition'")
        run_unconnected_check = run_unconnected_check or (
            has_definition and self._syntax_style_ is _SyntaxStyle.NEW)
        run_unconnected_check = run_unconnected_check and not \
            dct.get("_ignore_undriven_", False)
        if run_unconnected_check:
            with capture_logs(self._context_):
                find_and_log_unconnected_ports(self)

        return self

    @property
    def is_definition(self):
        return self._is_definition or self.verilog or self.verilogFile

    @property
    def instances(self):
        return self._context_.placer.instances()

    @property
    def logs(self):
        return self._context_.logs

    def place(cls, inst):
        """Place a circuit instance in this definition"""
        cls._context_.placer.place(inst)

    def bind(cls, monitor, *args, compile_guard=None):
        cls.bind_modules[monitor] = (args, compile_guard)


class Circuit(CircuitType, metaclass=DefineCircuitKind):
    _circuit_base_ = True


@deprecated(
    msg="DefineCircuit factory method is deprecated, subclass Circuit instead")
def DefineCircuit(name, *decl, **args):
    """DefineCircuit Factory"""
    debug_info = get_debug_info(4)
    metacls = DefineCircuitKind
    bases = (Circuit,)
    dct = metacls.__prepare__(name, bases)
    dct.update(dict(IO=decl,
                    is_definition=True,
                    primitive=args.get('primitive', False),
                    stateful=args.get('stateful', False),
                    simulate=args.get('simulate'),
                    debug_info=debug_info,
                    verilog_name=args.get('verilog_name', name),
                    coreir_name=args.get('coreir_name', name),
                    coreir_lib=args.get('coreir_lib', "global"),
                    coreir_genargs=args.get('coreir_genargs', None),
                    coreir_configargs=args.get('coreir_configargs', None),
                    default_kwargs=args.get('default_kwargs', {}),
                    renamed_ports=args.get('renamed_ports', {}),
                    kratos=args.get("kratos", None)))
    defn = metacls(name, bases, dct)
    push_definition_context(defn._context_)
    return defn


def EndDefine():
    # NOTE(rsetaluri): We first peek the definition context
    # (get_definition_context()) so that we avoid pushing on a log capturer for
    # the find_and_log_unconnected_ports() call.
    try:
        context = get_definition_context()
    except IndexError:
        raise Exception("EndDefine not matched to DefineCircuit")
    placer = context.placer
    find_and_log_unconnected_ports(placer._defn)
    debug_info = get_debug_info(3)
    placer._defn.end_circuit_filename = debug_info.filename
    placer._defn.end_circuit_lineno = debug_info.lineno
    pop_definition_context()


EndCircuit = EndDefine


def CopyInstance(instance):
    circuit = type(instance)
    new_instance = circuit()
    new_instance.kwargs = instance.kwargs
    return new_instance


_GeneratorArguments = namedtuple('GeneratorArguments', ['args', 'kwargs'])


@deprecated(msg="circuit_generator decorator is deprecated, subclass Generator "
            "instead")
def circuit_generator(func):
    @cache_definition
    @wraps(func)
    def wrapped(*args, **kwargs):
        result = func(*args, **kwargs)
        # Store arguments to generate the circuit.
        result._generator_arguments = _GeneratorArguments(args, kwargs)
        return result
    return wrapped


coreir_port_mapping = {
    "I": "in",
    "I0": "in0",
    "I1": "in1",
    "O": "out",
    "S": "sel",
    "CLK": "clk",
}


@deprecated
def DeclareCoreirCircuit(*args, **kwargs):
    return DeclareCircuit(*args, **kwargs,
                          renamed_ports=coreir_port_mapping)


def declare_coreir_circuit(name_: str, ports: dict, coreir_name_: str,
                           coreir_genargs_: dict, coreir_lib_: str):
    """
    parameter names are suffixed with `_` otherwise we get a NameError inside
    the class definition
    """
    class CoreIRCircuit(Circuit):
        name = name_
        renamed_ports = coreir_port_mapping
        io = IO(**ports)
        coreir_name = coreir_name_
        coreir_genargs = coreir_genargs_
        coreir_lib = coreir_lib_

    return CoreIRCircuit


class _CircuitBuilderMeta(type):
    pass


def builder_method(func):

    @wraps(func)
    def _wrapped(this, *args, **kwargs):
        with definition_context_manager(this.context):
            result = func(this, *args, **kwargs)
        return result

    return _wrapped


class CircuitBuilder(metaclass=_CircuitBuilderMeta):
    _RESERVED_NAMESPACE_KEYS = {"io", "_context_", "name"}

    def __init__(self, name):
        # Try to add this builder to a context if one exists currently. A
        # builder can still be constructed outside of a context, but (a) the
        # builder will not be automatically finalized, and (b) instantation
        # might fail.
        try:
            context = get_definition_context()
        except IndexError:
            pass
        else:
            context.add_builder(self)
        self._name = name
        self._io = SingletonInstanceIO()
        self._finalized = False
        self._dct = {}
        self._inst_attrs = {}
        self._context = DefinitionContext(StagedPlacer(self._name))
        self._instance_name = None

    def _port(self, name):
        return self._io.ports[name]

    def _add_port(self, name, typ):
        self._io.add(name, typ)
        setattr(self, name, self._io.inst_ports[name])
        return self._port(name)

    def _add_ports(self, **kwargs):
        return list(self._add_port(name, typ) for name, typ in kwargs.items())

    @property
    def _instances(self):
        return self._context.placer.instances.copy()

    def _set_definition_attr(self, key, value):
        if key in CircuitBuilder._RESERVED_NAMESPACE_KEYS:
            raise Exception(f"Can not set reserved attr '{key}'")
        self._dct[key] = value

    def _set_inst_attr(self, key, value):
        self._inst_attrs[key] = value

    def _open(self):
        return definition_context_manager(self._context)

    def _finalize(self):
        pass

    def set_instance_name(self, name):
        self._instance_name = name

    @property
    def context(self):
        return self._context

    @property
    def instance_name(self):
        if self._instance_name is None:
            return self._name
        return self._instance_name

    @property
    def defn(self):
        if not self._finalized:
            return None
        return self._defn

    def finalize(self, dont_instantiate: bool = False):
        if self._finalized:
            raise Exception("Can only call finalize on a CircuitBuilder once")
        self._finalize()
        bases = (AnonymousCircuitType,)
        dct = {"io": self._io, "_context_": self._context, "name": self._name}
        dct.update(self._dct)
        DefineCircuitKind.__prepare__(self._name, bases)
        self._defn = DefineCircuitKind(self._name, bases, dct)
        self._context.finalize(self._defn)
        self._finalized = True
        if dont_instantiate:
            return None
        inst = self._defn(name=self.instance_name)
        for k, v in self._inst_attrs.items():
            setattr(inst, k, v)
        return inst


class DebugDefineCircuitKind(DefineCircuitKind):
    def __prepare__(name, bases, **kwargs):
        prev_debug_mode = get_debug_mode()
        set_debug_mode(True)
        # NOTE(leonardt): Using super() here doesn't work:
        #   cls = super().__prepare__(name, bases, **kwargs)
        #   TypeError: super(type, obj): obj must be an instance or subtype of
        #   type
        cls = DefineCircuitKind.__prepare__(name, bases, **kwargs)
        ctx = get_definition_context()
        ctx.set_metadata("prev_debug_mode", prev_debug_mode)
        return cls

    def __new__(metacls, name, bases, dct):
        ctx = get_definition_context()
        set_debug_mode(ctx.get_metadata("prev_debug_mode"))
        return DefineCircuitKind.__new__(metacls, name, bases, dct)


class DebugCircuit(Circuit, metaclass=DebugDefineCircuitKind):
    pass


InstanceCallback = Callable[[List[CircuitType], Dict], None]


def register_instance_callback(
        instance: CircuitType,
        callback: InstanceCallback,
):
    if hasattr(instance, "_callback_"):
        raise AttributeError(
            f"Instance {instance} already has callback registered"
        )
    instance._callback_ = callback


def get_instance_callback(instance: CircuitType) -> InstanceCallback:
    try:
        return instance._callback_
    except AttributeError:
        msg = (
            f"Instance {instance} has no callback registered. Did you call "
            f"m.register_instance_callback()?"
        )
        raise AttributeError(msg) from None



from magma.array import Array
from magma.bit import Bit
from magma.bits import Bits
from magma.tuple import Tuple
from magma import clear_cachedFunctions

from magma.frontend import coreir_
from magma.generator import reset_generator_cache
from magma.logging import flush_all
from magma.when import reset_context as reset_when_context


def BitOrBits(width):
    if width is None:
        return Bit
    if not isinstance(width, int):
        raise ValueError(f"Expected width to be None or int, got {width}")
    return Bits[width]


def pretty_str(t):
    if issubclass(t, Tuple):
        args = []
        for i in range(len(t)):
            key_str = str(list(t.keys())[i])
            val_str = pretty_str(list(t.types())[i])
            indent = " " * 4
            val_str = f"\n{indent}".join(val_str.splitlines())
            args.append(f"{key_str} = {val_str}")
        # Pretty print by using newlines + indent
        joiner = ",\n    "
        result = joiner.join(args)
        # Insert first newline + indent and last newline
        result = "\n    " + result + "\n"
        s = f"Tuple({result})"
    elif issubclass(t, Bits):
        s = str(t)
    elif issubclass(t, Array):
        s = f"Array[{t.N}, {pretty_str(t.T)}]"
    else:
        s = str(t)
    return s


def pretty_print_type(t):
    print(pretty_str(t))


def reset_global_context():
    """
    Add this to your conftest.py to reset the magma compiler between tests

    Should also be used in between multiple calls of m.compile to reset the
    compiler's state
    """
    clear_cachedFunctions()
    coreir_.ResetCoreIR()
    reset_generator_cache()
    flush_all()  # flush all staged logs
    reset_when_context()



import itertools
from typing import Iterable, Mapping

from magma.array import Array
from magma.circuit import Circuit, DefineCircuitKind
from magma.clock import (
    ClockTypes, Clock, Reset, Enable, ResetN, AsyncReset, AsyncResetN,
    is_clock_or_nested_clock)
from magma.common import (
    only, IterableException, EmptyIterableException,
    NonSingletonIterableException)
from magma.logging import root_logger
from magma.passes.passes import DefinitionPass, pass_lambda
from magma.primitives.wire import Wire
from magma.t import Type, Kind
from magma.tuple import Tuple


_logger = root_logger()


def _get_output_clocks_in_array(
        value: Array, clock_type: Kind) -> Iterable[Type]:
    """
    Gets all output values of type @clock_type recursively contained in @value.
    """
    if not is_clock_or_nested_clock(type(value).T, (clock_type, )):
        return None
    for elem in value:
        yield from get_output_clocks_in_value(elem, clock_type)


def _get_output_clocks_in_tuple(
        value: Tuple, clock_type: Kind
) -> Iterable[Type]:
    """
    Gets all output values of type @clock_type recursively contained in @value.
    """
    for key, type_ in type(value).field_dict.items():
        if is_clock_or_nested_clock(type_, (clock_type, )):
            yield from get_output_clocks_in_value(value[key], clock_type)


def get_output_clocks_in_value(value: Type, clock_type: Kind) -> Iterable[Type]:
    """
    Gets all output values of type @clock_type recursively contained in @value.
    """
    # Since we are looking for default drivers, only need to consider outputs.
    if value.is_input():
        return None
    if isinstance(value, clock_type):
        yield value
    if isinstance(value, Tuple):
        yield from _get_output_clocks_in_tuple(value, clock_type)
    if isinstance(value, Array):
        yield from _get_output_clocks_in_array(value, clock_type)


def get_output_clocks_in_defn(
        defn: DefineCircuitKind, clock_type: Kind) -> Iterable[Type]:
    """
    Gets all output values of type @clock_type in circuit @defn. This includes
    "input" ports of @defn and output ports of any instances contained in @defn.
    """
    for port in defn.interface.ports.values():
        yield from get_output_clocks_in_value(port, clock_type)
    for inst in defn.instances:
        if isinstance(type(inst), Wire):
            # We want to skip instances of Wire(T), because they are basically
            # buffers and therefore their outputs are not legitimate
            # drivers. Not doing so might result in spurrious "multiple clocks".
            continue
        for port in inst.interface.ports.values():
            yield from get_output_clocks_in_value(port, clock_type)


def get_all_output_clocks_in_defn(
        defn: DefineCircuitKind) -> Mapping[Kind, Type]:
    """
    Returns a map from each clock type to either the singleton output value of
    that type in @defn, or None (if there is none, or there are multiple).
    """
    default_clocks = {}
    for clock_type in ClockTypes:
        clocks = get_output_clocks_in_defn(defn, clock_type)
        try:
            clock = only(clocks)
        except IterableException:
            clock = None
        default_clocks[clock_type] = clock
    return default_clocks


def _get_undriven_clocks_in_array(
        value: Array, clock_type: Kind) -> Iterable[Type]:
    """
    Returns all undriven values of type @clock_type contained in array @value.
    """
    # First check if the first element has any undriven.
    it = get_undriven_clocks_in_value(value[0], clock_type)
    try:
        first = next(it)
    except StopIteration:  # early exit to avoid traversing other elements
        return
    yield first
    yield from it
    # NOTE(rsetaluri): Since magma does not allow zero-length arrays, we need to
    # check that the length of the array is > 1 before slicing it with [1:].
    if len(value) <= 1:
        return
    for elem in value[1:]:
        yield from get_undriven_clocks_in_value(elem, clock_type)


def get_undriven_clocks_in_value(
        value: Type, clock_type: Kind) -> Iterable[Type]:
    """Returns all undriven values of type @clock_type contained in @value."""
    if not is_clock_or_nested_clock(type(value), (clock_type, )):
        # Avoid descening into tuple/array if possible
        return
    if isinstance(value, Tuple):
        for elem in value:
            yield from get_undriven_clocks_in_value(elem, clock_type)
        return
    if isinstance(value, Array):
        yield from _get_undriven_clocks_in_array(value, clock_type)
        return
    if isinstance(value, clock_type) and value.trace() is None:
        # Trace back to last undriven driver.
        while value.driven():
            value = value.value()
        yield value
    return


def _drive_value_with_clock(value: Type, clk: Type):
    _logger.debug(f"Auto-wiring {repr(clk)} to {repr(value)}")
    value @= clk


def drive_undriven_clocks_in_inst(
        defn: DefineCircuitKind, inst: Circuit, clock_type: Kind):
    """
    Drives all undriven input ports of @inst of type @clock_type with a default
    driver of type @clock_type contained in @defn. If no such default driver
    exists (either due to no drivers or multiple drivers existing), then this
    function is a no-op.
    """
    undrivens = iter(())
    for port in inst.interface.inputs(include_clocks=True):
        undrivens = itertools.chain(
            undrivens, get_undriven_clocks_in_value(port, clock_type))

    try:
        undriven = next(undrivens)
    except StopIteration:
        return
    clks = get_output_clocks_in_defn(defn, clock_type)
    try:
        clk = only(clks)
    except EmptyIterableException:
        _logger.warning(
            f"Found no clocks in {defn.name}; skipping auto-wiring "
            f"{clock_type}")
        return
    except NonSingletonIterableException as e:
        _logger.warning(
            f"Found multiple clocks in {defn.name}; skipping auto-wiring "
            f"{clock_type} ({', '.join(map(repr, e.args[0]))})"
        )
        return
    # Restore undrivens after popping off the first element.
    undrivens = itertools.chain([undriven], undrivens)
    for undriven in undrivens:
        _drive_value_with_clock(undriven, clk)


def drive_all_undriven_clocks_in_value(
        value: Type, clocks: Mapping[Kind, Type]) -> bool:
    """
    Drives all undriven input values contained in @value for each of the (clock
    type, default driver) pairs in @clocks.
    """
    has_undriven = False
    for clock_type, clk in clocks.items():
        if clk is None:
            continue
        undrivens = get_undriven_clocks_in_value(value, clock_type)
        for undriven in undrivens:
            has_undriven = True
            _drive_value_with_clock(undriven, clk)
    return has_undriven


def drive_undriven_clock_types_in_inst(defn: DefineCircuitKind, inst: Circuit):
    drive_undriven_clocks_in_inst(defn, inst, Clock)


def drive_undriven_other_clock_types_in_inst(
        defn: DefineCircuitKind, inst: Circuit):
    drive_undriven_clocks_in_inst(defn, inst, Reset)
    drive_undriven_clocks_in_inst(defn, inst, ResetN)
    drive_undriven_clocks_in_inst(defn, inst, AsyncReset)
    drive_undriven_clocks_in_inst(defn, inst, AsyncResetN)
    drive_undriven_clocks_in_inst(defn, inst, Enable)


class WireClockPass(DefinitionPass):
    def __call__(self, definition):
        with definition.open():
            for instance in definition.instances:
                drive_undriven_clock_types_in_inst(definition, instance)
                drive_undriven_other_clock_types_in_inst(definition, instance)


wire_clocks = pass_lambda(WireClockPass)



from magma.frontend.verilog_importer import VerilogImporter


class DummyVerilogImporter(VerilogImporter):
    def __init__(self):
        super().__init__({})

    def import_(self, src, mode):
        raise NotImplementedError()



try:
    import matplotlib.pyplot as plt
    import numpy as np

    def waveform(signals, labels):

        signals = np.transpose(np.array(signals))

        n = len(signals[0])

        nsignals = len(signals)

        t = np.repeat(0.5*np.arange(n+1),2)[1:-1]
        signals = [np.repeat(s,2) for s in signals]

        for i in range(nsignals):
            plt.text(-1.5, i+.25, labels[i])
            plt.plot(t, 0.5*signals[i]+i, 'r', linewidth = 2)

        # plt.xlim([0,n])
        # plt.ylim([-.25,nsignals])

        plt.gca().axis('off')
        plt.show()
except ImportError as e:
    def waveform(signals, labels):
        raise Exception("Please install matplotlib to use this feature.")



import dataclasses
from typing import List

from magma.backend.mlir.mlir import MlirDialect, begin_dialect, end_dialect
from magma.backend.mlir.mlir_printer_utils import print_names, print_types
from magma.backend.mlir.mlir import MlirValue, MlirOp
from magma.backend.mlir.printer_base import PrinterBase


comb = MlirDialect("comb")
begin_dialect(comb)


@dataclasses.dataclass
class BaseOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]
    op_name: str

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(f" = comb.{self.op_name} ")
        print_names(self.operands, printer)
        printer.print(" : ")
        print_types(self.results, printer)


@dataclasses.dataclass
class ConcatOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(f" = comb.concat ")
        print_names(self.operands, printer)
        printer.print(" : ")
        print_types(self.operands, printer)


@dataclasses.dataclass
class ExtractOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]
    lo: int

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(" = comb.extract ")
        print_names(self.operands, printer)
        printer.print(f" from {self.lo} : (")
        print_types(self.operands, printer)
        printer.print(") -> ")
        print_types(self.results, printer)


@dataclasses.dataclass
class ICmpOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]
    predicate: str

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(f" = comb.icmp {self.predicate} ")
        print_names(self.operands, printer)
        printer.print(f" : ")
        print_types(self.operands[0], printer)


@dataclasses.dataclass
class ParityOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(f" = comb.parity ")
        print_names(self.operands, printer)
        printer.print(" : ")
        print_types(self.operands, printer)


end_dialect()



import dataclasses
import functools
import itertools
from typing import Optional

from magma.clock import is_clock_or_nested_clock, ClockTypes
from magma.compile_exception import make_unconnected_error_str
from magma.debug import debug_info as debug_info_cls
from magma.logging import root_logger
from magma.t import Type
from magma.wire_container import WiringLog


_logger = root_logger()


@dataclasses.dataclass(frozen=True)
class UnconnectedPortInfo:
    port: Type
    debug_info: Optional[debug_info_cls]


def find_unconnected_ports_of_port(port, debug_info):
    if port.is_output():
        return
    if port.is_inout() and port.wired():
        # TODO(leonardt/rsetaluri): Ideally, we should trace in both directions
        # for an inout; for now we assume that if it's "wired()", then it is
        # connected.
        return
    if port.is_mixed():
        for elt in port:
            yield from find_unconnected_ports_of_port(elt, debug_info)
        return
    if port.trace() is None:
        # If we encounter a clock, issue a debug message that we are not
        # counting this as an unconnected port and instead will attempt to
        # automatically wire it downstream.
        if isinstance(port, ClockTypes):
            msg = "{} not driven, will attempt to automatically wire"
            _logger.debug(WiringLog(msg, port), debug_info=debug_info)
            return
        # If there is a clock nested in @port, then we need to traverse its
        # children. Note that although we use is_clock_or_nested_clock here,
        # because of the preceding check for ClockTypes, this conditional will
        # only catch the nested clock case.
        if is_clock_or_nested_clock(type(port)):
            for elt in port:
                yield from find_unconnected_ports_of_port(elt, debug_info)
            return
        yield UnconnectedPortInfo(port, debug_info)
        return
    return


def find_unconnected_ports_of_interface(interface, debug_info):
    for port in interface.ports.values():
        yield from find_unconnected_ports_of_port(port, debug_info)


def find_unconnected_ports_of_circuit(ckt):
    yield from find_unconnected_ports_of_interface(
        ckt.interface, ckt.debug_info
    )
    for inst in ckt.instances:
        yield from find_unconnected_ports_of_interface(
            inst.interface, inst.debug_info
        )


@dataclasses.dataclass
class UnconnectedPortDiagnostic:
    value: Type
    connected: Optional[bool]
    depth: int

    def make_wiring_log(self) -> str:
        tab = " " * (4 * self.depth)
        connected_str = ""
        if self.connected is True:
            connected_str = ": Connected"
        elif self.connected is False:
            connected_str = ": Unconnected"
        tpl = f"{tab}{{}}{connected_str}"
        return WiringLog(tpl, self.value)


# NOTE(rsetaluri): The value visitor class for constructing the unconnected port
# diagnostic info requires inheriting from ValueVisitor (defined in
# magma/value_utils.py). This would result in a cyclical import dependency. To
# hack around this issue, the value visitor class is created in a factory which
# should only run once, and use the cached result for future invocations. Note
# that we set @maxsize to 1 for the lru cache.
@functools.lru_cache(maxsize=1)
def _make_unconnected_port_diagnostic_visitor_cls():
    from magma.value_utils import ValueVisitor

    class _Visitor(ValueVisitor):
        def __init__(self):
            self._depth = 0

        def visit(self, node):
            yield from super().visit(node)

        def generic_visit(self, value):
            if value.trace() is not None:
                yield UnconnectedPortDiagnostic(value, True, self._depth)
                return
            wrapped = self.wrap(value)
            results = []
            has_connected = False
            self._depth += 1
            for child in wrapped.children:
                for diagnostic in self.visit(child):
                    has_connected = has_connected or diagnostic.connected
                    results.append(diagnostic)
            self._depth -= 1
            if not has_connected:
                yield UnconnectedPortDiagnostic(value, False, self._depth)
                return
            yield UnconnectedPortDiagnostic(value, None, self._depth)
            yield from iter(results)

    return _Visitor


def find_and_log_unconnected_ports(ckt):
    infos = find_unconnected_ports_of_circuit(ckt)
    for info in infos:
        port = info.port
        debug_info = info.debug_info
        _logger.error(WiringLog("{} not driven", port), debug_info=debug_info)
        visitor = _make_unconnected_port_diagnostic_visitor_cls()()
        diagnostics = visitor.visit(port)
        for diagnostic in diagnostics:
            _logger.error(diagnostic.make_wiring_log())



import abc
import dataclasses
import inspect
import operator
from typing import Callable, Union

from magma.bit import Bit
from magma.bits import Bits, BitsMeta, SInt, reduce as bits_reduce
from magma.conversions import uint, bits, sint
from magma.conversions import concat as bits_concat
from magma.debug import debug_wire
from magma.protocol_type import MagmaProtocolMeta, MagmaProtocol
from magma.ref import AnonRef


class SmartExprMeta(MagmaProtocolMeta):
    pass


class SmartExpr(MagmaProtocol, metaclass=SmartExprMeta):
    __hash__ = object.__hash__

    def __init__(self):
        self._name = AnonRef()
        self._value = None

    @property
    def name(self):
        if self._value is None:
            return self._name
        return self._value.name

    @name.setter
    def name(self, value):
        if self._value is None:
            self._name = value
        else:
            self._value.name = value

    @property
    @abc.abstractmethod
    def args(self):
        raise NotImplementedError()

    # Binary arithmetic operators.
    def __add__(self, other: 'SmartExpr') -> 'SmartBinaryOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartBinaryOp(operator.add, self, other)

    def __sub__(self, other: 'SmartExpr') -> 'SmartBinaryOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartBinaryOp(operator.sub, self, other)

    def __mul__(self, other: 'SmartExpr') -> 'SmartBinaryOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartBinaryOp(operator.mul, self, other)

    def __floordiv__(self, other: 'SmartExpr') -> 'SmartBinaryOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartBinaryOp(operator.floordiv, self, other)

    def __truediv__(self, other: 'SmartExpr') -> 'SmartBinaryOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartBinaryOp(operator.truediv, self, other)

    def __mod__(self, other: 'SmartExpr') -> 'SmartBinaryOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartBinaryOp(operator.mod, self, other)

    # Binary logic operators.
    def __and__(self, other: 'SmartExpr') -> 'SmartBinaryOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartBinaryOp(operator.and_, self, other)

    def __or__(self, other: 'SmartExpr') -> 'SmartBinaryOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartBinaryOp(operator.or_, self, other)

    def __xor__(self, other: 'SmartExpr') -> 'SmartBinaryOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartBinaryOp(operator.xor, self, other)

    # Comparison operators.
    def __eq__(self, other: 'SmartExpr') -> 'SmartComparisonOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartComparisonOp(operator.eq, self, other)

    def __ne__(self, other: 'SmartExpr') -> 'SmartComparisonOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartComparisonOp(operator.ne, self, other)

    def __ge__(self, other: 'SmartExpr') -> 'SmartComparisonOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartComparisonOp(operator.ge, self, other)

    def __gt__(self, other: 'SmartExpr') -> 'SmartComparisonOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartComparisonOp(operator.gt, self, other)

    def __le__(self, other: 'SmartExpr') -> 'SmartComparisonOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartComparisonOp(operator.le, self, other)

    def __lt__(self, other: 'SmartExpr') -> 'SmartComparisonOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartComparisonOp(operator.lt, self, other)

    # Shift operators.
    def __lshift__(self, other: 'SmartExpr') -> 'SmartShiftOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartShiftOp(operator.lshift, self, other)

    def __rshift__(self, other: 'SmartExpr') -> 'SmartShiftOp':
        if not isinstance(other, SmartExpr):
            return NotImplemented
        return SmartShiftOp(operator.rshift, self, other)

    # Unary operators.
    def __invert__(self) -> 'SmartUnaryOp':
        return SmartUnaryOp(operator.invert, self)

    def __neg__(self) -> 'SmartUnaryOp':
        return SmartUnaryOp(operator.neg, self)

    # Reduction operators.
    def reduce(self, op) -> 'SmartReductionOp':
        return SmartReductionOp(op, self)

    def reduce_and(self) -> 'SmartReductionOp':
        return self.reduce(operator.and_)

    def reduce_or(self) -> 'SmartReductionOp':
        return self.reduce(operator.or_)

    def reduce_xor(self) -> 'SmartReductionOp':
        return self.reduce(operator.xor)

    # Extension operators.
    def zext(self, width) -> 'SmartExtendOp':
        return SmartExtendOp(width, False, self)

    def sext(self, width) -> 'SmartExtendOp':
        return SmartExtendOp(width, True, self)

    def const(self) -> bool:
        return False


class SmartOp(SmartExpr, metaclass=SmartExprMeta):
    def __init__(self, op, *args):
        super().__init__()
        self._op = op
        self._args = args

    @property
    def op(self):
        return self._op

    @property
    def args(self):
        return self._args

    def __str__(self):
        op = self.op.__name__ if inspect.isbuiltin(self.op) else str(self.op)
        args = ", ".join(str(arg) for arg in self.args)
        return f"{op}({args})"


class SmartExtendOp(SmartOp):

    @dataclasses.dataclass(frozen=True)
    class Op:
        width: int
        signed: bool

        def __call__(self, operand: Bits) -> Bits:
            if self.signed:
                return sint(operand).sext(self.width)
            return uint(operand).zext(self.width)

        def __str__(self):
            func = "Sext" if self.signed else "Zext"
            return f"{func}[{self.width}]"

    def __init__(self, width, signed, operand):
        super().__init__(SmartExtendOp.Op(width, signed), operand)


class SmartReductionOp(SmartOp):
    OP_TO_NAME = {
        operator.and_: "And",
        operator.or_: "Or",
        operator.xor: "Xor",
    }

    @dataclasses.dataclass(frozen=True)
    class Op:
        op: Callable

        def __post_init__(self):
            if self.op not in SmartReductionOp.OP_TO_NAME:
                raise ValueError(f"Reduction operator {op} not supported")

        def __call__(self, operand: Bits) -> Bits:
            return bits(bits_reduce(self.op, operand))

        def __str__(self):
            func = SmartReductionOp.OP_TO_NAME[self.op]
            return f"{func}Reduce"

    def __init__(self, op, operand):
        super().__init__(SmartReductionOp.Op(op), operand)


class SmartNAryContextualOp(SmartOp):
    pass


class SmartBinaryOp(SmartNAryContextualOp):
    def __init__(self, op, loperand, roperand):
        super().__init__(op, loperand, roperand)


class SmartUnaryOp(SmartNAryContextualOp):
    def __init__(self, op, operand):
        super().__init__(op, operand)


class SmartComparisonOp(SmartOp):

    @dataclasses.dataclass(frozen=True)
    class Op:
        op: Callable

        def __call__(self, *args) -> Bits:
            return bits(self.op(*args))

        def __str__(self) -> str:
            return self.op.__name__

    def __init__(self, op, loperand, roperand):
        super().__init__(SmartComparisonOp.Op(op), loperand, roperand)


class SmartShiftOp(SmartOp):
    def __init__(self, op, loperand, roperand):
        super().__init__(op, loperand, roperand)


class SmartConcatOp(SmartOp):

    class Op:
        def __call__(self, *args) -> Bits:
            return bits_concat(*args)

        def __str__(self) -> str:
            return "Concat"

    def __init__(self, *args):
        super().__init__(SmartConcatOp.Op(), *args)


class SmartMuxOp(SmartOp):

    class Op:
        def __call__(self, *args) -> Bits:
            import magma as m
            values, select = args[:-1], args[-1]
            return m.mux(values, select)

        def __str__(self) -> str:
            return "Mux"

    def __init__(self, *args):
        super().__init__(SmartMuxOp.Op(), *args)


def concat(*args) -> SmartExpr:
    if not all(isinstance(arg, SmartExpr) for arg in args):
        types = ", ".join(str(type(arg)) for arg in args)
        raise NotImplementedError(f"Concat not supported for [{types}]")
    return SmartConcatOp(*args)


def repeat(value, num) -> SmartExpr:
    """Repeats @value @num number of times flat. Note that for single bit values
    (e.g. x[0]) this is the same as m.repeat, but for bits this function retruns
    a flattened type, whereas m.repeat constructs a higher-dimensional
    array. This function is equivalent to m.smart.concat(*(value for _ in
    range(num))).
    """
    return concat(*(value for _ in range(num)))


def mux(values, select) -> SmartExpr:
    if not all(isinstance(value, SmartExpr) for value in values):
        types = ", ".join(str(type(value)) for value in values)
        raise NotImplementedError(f"Mux not supported for [{types}]")
    return SmartMuxOp(*values, select)


class SmartSignedOp(SmartOp):

    @dataclasses.dataclass
    class Op:
        signed: bool

        def __call__(self, operand: Bits) -> Bits:
            cons = sint if self.signed else uint
            return cons(operand)

        def __str__(self):
            return "Signed" if self.signed else "Unsigned"

    def __init__(self, signed, operand):
        super().__init__(SmartSignedOp.Op(signed), operand)


def signed(expr):
    return SmartSignedOp(True, expr)


def unsigned(expr):
    return SmartSignedOp(False, expr)


@dataclasses.dataclass(frozen=True)
class SmartBitsExpr(SmartExpr, metaclass=SmartExprMeta):
    bits: 'SmartBits'

    @property
    def args(self):
        return []

    def __str__(self):
        return str(self.bits)


class SmartBitsMeta(SmartExprMeta):
    def _parse_key(cls, key):
        qualified = (
            issubclass(cls, SmartBits) and
            (
                hasattr(cls, "_T_") or
                hasattr(cls, "_signed_")
            )
        )
        if qualified:
            raise TypeError("Can not doubly qualify SmartBits")
        width_or_type = key
        signed = False
        if isinstance(key, tuple):
            if len(key) == 1:
                width_or_type = key[0]
            elif len(key) == 2:
                width_or_type, signed = key
            else:
                raise ValueError(f"{key} unsupported")
        try:
            int(width_or_type)
        except (ValueError, TypeError):
            assert isinstance(width_or_type, BitsMeta)
            T = width_or_type
        else:
            T = Bits[width_or_type]
        return T, signed

    def __eq__(cls, other):
        return (
            isinstance(other, SmartExprMeta) and
            cls._to_magma_() == other._to_magma_()
        )

    __hash__ = type.__hash__

    def __getitem__(cls, key):
        T, signed = SmartBitsMeta._parse_key(cls, key)
        name = f"SmartBits[{len(T)}, {signed}]"
        dct = {"_T_": T, "_signed_": signed}
        return type(cls)(name, (cls,), dct)

    def _to_magma_(cls):
        return cls._T_

    def _qualify_magma_(cls, d):
        return SmartBits[cls._T_.qualify(d), cls._signed_]

    def _flip_magma_(cls):
        return SmartBits[cls._T_.flip(), cls._signed_]

    def _from_magma_value_(cls, value):
        return cls(value)

    def __repr__(cls):
        has_T = hasattr(cls, "_T_")
        has_signed = hasattr(cls, "_signed_")
        if not (has_T or has_signed):
            return "SmartBits"
        assert has_T and has_signed
        return f"SmartBits[{len(cls._T_)}, {cls._signed_}]"


class SmartBits(SmartBitsExpr, metaclass=SmartBitsMeta):
    def __init__(self, value=None, name=None):
        super().__init__(self)
        if value is None:
            value = type(self)._to_magma_()(name=name)
        self._value = value

    def __len__(self):
        return len(type(self)._T_)

    def typed_value(self):
        if type(self)._signed_:
            return sint(self._value)
        return uint(self._value)

    def untyped_value(self):
        return self._value

    def __hash__(self):
        return hash(self._value)

    def _get_magma_value_(self):
        return self.untyped_value()

    # Slice operators.
    def __getitem__(self, key_or_slice) -> 'SmartBits':
        return SmartBits.from_bits(self._value[key_or_slice])

    @debug_wire
    def wire(self, other, debug_info):
        if isinstance(other, Bits):
            MagmaProtocol.wire(self, other, debug_info)
            return
        if not isinstance(other, SmartExpr):
            raise ValueError(f"Can not wire {type(self)} to {type(other)}")
        # NOTE(rsetaluri): We delay the import of the evaluation method to avoid
        # a circular import, since the evaluation depends on the implementation
        # of SmartBits.
        from magma.smart.eval import evaluate_assignment
        other = evaluate_assignment(self, other)
        MagmaProtocol.wire(self, other, debug_info)

    @staticmethod
    def from_bits(value):
        if isinstance(value, Bit):
            value = bits(value)
        if not isinstance(value, Bits):
            raise TypeError(value)
        signed = isinstance(value, SInt)
        return SmartBits[len(value), signed](value)

    def __str__(self):
        signed = type(self)._signed_
        return f"SmartBits[{len(self)}, {signed}]({str(self._value)})"

    def connection_iter(self):
        yield from zip(self, self.trace())


SmartBit = SmartBits[1]


def issigned(value_or_type: Union[SmartBits, SmartBitsMeta]) -> bool:
    if isinstance(value_or_type, SmartBits):
        return type(value_or_type)._signed_
    if isinstance(value_or_type, SmartBitsMeta):
        return value_or_type._signed_
    raise TypeError(value_or_type)


def eval(expr: SmartExpr, width: int, signed: bool = False):
    lhs = SmartBits[width, signed]()
    lhs @= expr
    return lhs



from magma.passes.passes import DefinitionPass, pass_lambda
from magma.type_utils import contains_tuple


def _expand(value):
    if contains_tuple(type(value)):
        for elem in value:
            _expand(elem)


def _expand_tuple_values(values):
    for value in values:
        if contains_tuple(type(value)):
            _expand(value)


class ElaborateTuplesPass(DefinitionPass):
    def __call__(self, definition):
        with definition.open():
            _expand_tuple_values(definition.interface.ports.values())
            for instance in definition.instances:
                _expand_tuple_values(instance.interface.ports.values())


elaborate_tuples = pass_lambda(ElaborateTuplesPass)



import dataclasses
import typing

from magma.ref import Ref, TupleRef, ArrayRef
from magma.type_utils import (isprotocol, isdigital, isbits, isarray, istuple,
                              GenericWrapper, WrappedVisitor)


def _unwrap_protocol_value(value):
    T = type(value)
    if isprotocol(T):
        return value._get_magma_value_()
    return value


@dataclasses.dataclass(frozen=True)
class _ValueWrapper(GenericWrapper):
    pass


def _wrap_value(value):
    if isinstance(value, _ValueWrapper):
        return value
    value = _unwrap_protocol_value(value)
    T = type(value)

    if isdigital(T):
        return _ValueWrapper(value, [], "Digital")
    if isbits(T):
        return _ValueWrapper(value, [t for t in value], "Bits")
    if isarray(T):
        return _ValueWrapper(value, [t for t in value], "Array")
    if istuple(T):
        return _ValueWrapper(value, [t for t in value], "Tuple")

    raise NotImplementedError(T)


class ValueVisitor(WrappedVisitor):
    def wrap(self, value):
        return _wrap_value(value)


class ValueTransformer(ValueVisitor):
    def generic_visit(self, value):
        wrapped = _wrap_value(value)
        args = (self.visit(child) for child in wrapped.children)
        return type(value)(*args)


@dataclasses.dataclass(frozen=True)
class Selector:
    child: 'Selector'

    def select(self, value):
        value = self._select(value)
        if self.child is None:
            return value
        return self.child.select(value)

    def _select(self, value):
        return value

    def _child_str(self):
        return "" if self.child is None else str(self.child)


@dataclasses.dataclass(frozen=True)
class TupleSelector(Selector):
    key: str

    def _select(self, value):
        return getattr(value, self.key)

    def __str__(self):
        return f".{self.key}{self._child_str()}"


@dataclasses.dataclass(frozen=True)
class ArraySelector(Selector):
    index: int

    def _select(self, value):
        return value[self.index]

    def __str__(self):
        return f"[{self.index}]{self._child_str()}"


def _make_selector_impl(value, child, stop_at):
    value = _unwrap_protocol_value(value)
    if value is not stop_at:
        ref = value.name
        if isinstance(ref, ArrayRef):
            child = ArraySelector(child, ref.index)
            return _make_selector_impl(ref.array, child, stop_at)
        if isinstance(ref, TupleRef):
            child = TupleSelector(child, ref.index)
            return _make_selector_impl(ref.tuple, child, stop_at)
    if child is not None:
        return child
    return Selector(None)


def make_selector(value, stop_at=None):
    """stop_at (optional): indicate parent value to 'stop at', useful if you
    want a partial selector of a child"""
    return _make_selector_impl(value, None, stop_at)


class _FillVisitor(ValueVisitor):
    def __init__(self, fill_value: bool):
        self._fill_value = fill_value

    def visit_Digital(self, value):
        value @= self._fill_value

    def visit_Bits(self, value):
        size = len(value)
        fill_value = 0 if not self._fill_value else (1 << size) - 1
        value @= fill_value


def fill(value, fill_value: bool):
    _FillVisitor(fill_value).visit(value)



from magma.circuit import CircuitKind
from magma.passes.passes import CircuitPass, pass_lambda


def elaborate_circuit(ckt: CircuitKind):
    with ckt.open():
        ckt.elaborate()
    if getattr(ckt, "primitive", False):
        setattr(ckt, "primitive", False)
    ckt._is_definition = True


class ElaborationPass(CircuitPass):
    def __init__(self, main):
        super().__init__(main)

    def should_elaborate(self, ckt):
        raise NotImplementedError()

    def __call__(self, ckt):
        if self.should_elaborate(ckt):
            elaborate_circuit(ckt)


class ElaborateAllPass(ElaborationPass):
    def __init__(self, main, circuits=(), generators=()):
        super().__init__(main)
        self._circuits = circuits
        self._generators = generators

    def should_elaborate(self, ckt):
        # NOTE(rsetaluri): We are using a subclass check for circuits. This is
        # not necessarily guaranteed to be equivalent to checking that the
        # instance itself isinstance of the circuit, but we can expect it to
        # work for most cases.
        return (
            issubclass(ckt, self._circuits)
            or isinstance(ckt, self._generators)
        )


elaborate_all_pass = pass_lambda(ElaborateAllPass)



import abc
import typing
import weakref
from magma.compatibility import IntegerTypes


class Ref:
    @abc.abstractmethod
    def __str__(self):
        raise NotImplementedError()

    def __repr__(self):
        return self.qualifiedname()

    @abc.abstractmethod
    def qualifiedname(self, sep="."):
        raise NotImplementedError()

    @abc.abstractmethod
    def anon(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def named(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def bound(self) -> bool:
        raise NotImplementedError()

    def parent(self):
        return self

    def root(self) -> typing.Optional['Ref']:
        parent = self.parent()
        if parent is self:
            return self
        return parent.root()


class AnonRef(Ref):
    def __init__(self):
        self.name = None

    def __str__(self):
        return f"AnonymousValue_{id(self)}"

    def qualifiedname(self, sep='.'):
        return f"AnonymousValue_{id(self)}"

    def anon(self):
        return True

    def named(self) -> bool:
        return False

    def bound(self) -> bool:
        return False


class ConstRef(Ref):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    def qualifiedname(self, sep="."):
        return self.name

    def anon(self):
        return False

    def named(self) -> bool:
        return False

    def bound(self) -> bool:
        return False


class NamedRef(Ref):
    def __init__(self, name, value=None):
        if not isinstance(name, (str, int)):
            raise TypeError("Expected string or int")
        self.name = name
        self._value = value if value is None else weakref.ref(value)

    def __str__(self):
        return self.name

    def qualifiedname(self, sep="."):
        return self.name

    def anon(self):
        return False

    def value(self):
        return self._value if self._value is None else self._value()

    def named(self) -> bool:
        return True


class TempNamedRef(NamedRef):
    def bound(self) -> bool:
        return False


class InstRef(NamedRef):
    def __init__(self, inst, name):
        super().__init__(name)
        if not inst:
            raise ValueError(f"Bad inst: {inst}")
        self.inst = inst

    def qualifiedname(self, sep="."):
        name = self.name
        if isinstance(self.name, IntegerTypes):
            # Hack, Hack, Hack!
            # NOTE: This is used for verilog instances that don't use named
            # port (wired by index instead), so the ports are referred to by
            # index instead of name and we use the array indexing syntax to
            # represent them
            # See mantle's generic verilog target for example use case
            if sep == ".":
                return f"{self.inst.name}[{self.name}]"
        return self.inst.name + sep + str(name)

    def bound(self) -> bool:
        return True


class LazyInstRef(InstRef):
    def __init__(self, name):
        self.name = name
        self._inst = None

    @property
    def inst(self):
        if self._inst is not None:
            return self._inst
        return LazyCircuit()

    def qualifiedname(self, sep="."):
        return super().qualifiedname(sep)

    def set_inst(self, inst):
        if self._inst is not None:
            raise Exception("Can only set definition of LazyInstRef once")
        self._inst = inst


class DefnRef(NamedRef):
    def __init__(self, defn, name):
        super().__init__(name)
        if not defn:
            raise ValueError(f"Bad defn: {defn}")
        self.defn = defn

    def qualifiedname(self, sep="."):
        if sep == ".":
            return self.defn.__name__ + sep + self.name
        return self.name

    def bound(self) -> bool:
        return True


class LazyCircuit:
    name = ""
    debug_name = ""

    @property
    def defn(self):
        return LazyCircuit


class LazyDefnRef(DefnRef):
    def __init__(self, name):
        self.name = name
        self._defn = None

    @property
    def defn(self):
        if self._defn is not None:
            return self._defn
        return LazyCircuit

    def qualifiedname(self, sep="."):
        return super().qualifiedname(sep)

    def set_defn(self, defn):
        if self._defn is not None:
            raise Exception("Can only set definition of LazyDefnRef once")
        self._defn = defn


class DerivedRef(Ref):
    def __init__(self, parent, index):
        self._parent = parent
        self._index = index

    def __str__(self):
        return self.qualifiedname()

    def anon(self) -> bool:
        return self.parent().anon()

    def named(self) -> bool:
        return self.parent().named()

    def bound(self) -> bool:
        return self.parent().bound()

    def parent(self):
        return self._parent.name

    @property
    def parent_value(self):
        return self._parent

    @property
    def index(self):
        return self._index


class ArrayRef(DerivedRef):
    def __init__(self, array, index):
        super().__init__(array, index)
        self.array = array

    def qualifiedname(self, sep="."):
        return f"{self.parent().qualifiedname(sep=sep)}[{self.index}]"


class TupleRef(DerivedRef):
    def __init__(self, tuple, index):
        super().__init__(tuple, index)
        self.tuple = tuple

    def qualifiedname(self, sep="."):
        try:
            int(self.index)
            return (self.parent().qualifiedname(sep=sep) +
                    "[" + str(self.index) + "]")
        except ValueError:
            return (self.parent().qualifiedname(sep=sep) +
                    sep + str(self.index))


class PortViewRef(Ref):
    """
    Used for values that are connection references to a hierarchical value
    (using the view logic)
    """

    def __init__(self, view):
        self.view = view

    def qualifiedname(self, sep="."):
        return self.view.port.name.qualifiedname(sep)

    def anon(self):
        return self.view.port.anon()

    def __str__(self):
        return str(self.view.port.name)

    def root(self):
        return self.view.root()


def get_ref_inst(ref):
    """
    If value is part of a port on an instance, return that instance,
    otherwise None.
    """
    root = ref.root()
    if not isinstance(root, InstRef):
        return None
    return root.inst


def get_ref_defn(ref):
    """
    If value is part of a port on an definition, return that definition,
    otherwise None.
    """
    root = ref.root()
    if not isinstance(root, DefnRef):
        return None
    return root.defn


def is_temp_ref(ref):
    root = ref.root()
    return isinstance(root, (TempNamedRef, AnonRef))


def get_parent_array(value):
    if not isinstance(value.name, ArrayRef):
        return None
    root_value = value
    while isinstance(root_value.name, ArrayRef):
        root_value = root_value.name.parent_value
    return root_value



from magma.backend.coreir.coreir_utils import Slice
from magma.ref import PortViewRef, get_ref_inst, get_ref_defn, is_temp_ref
from magma.protocol_type import magma_value
from magma.compile_exception import MagmaCompileException
from magma.view import InstView


def _get_inst(value):
    if isinstance(value, InstView):
        return value.inst
    return get_ref_inst(value.name)


def _get_defn(value):
    if isinstance(value, InstView):
        return None
    return get_ref_defn(value.name)


def check_wiring_context(i, o):
    """
    Ensures that i and o come from the same definition context
    """
    orig_i, orig_o = i, o
    if isinstance(o, Slice):
        o = o.value
    if isinstance(i, Slice):
        i = i.value
    i, o = magma_value(i), magma_value(o)
    if is_temp_ref(o.name) or is_temp_ref(i.name):
        # Will be handled recursively by the get_source logic for the coreir
        # backend, i.e. when a connection is made, we only ever use the source
        # and sink (temporary values are for non-whole arrays that we then
        # recurse over)
        # NOTE: For other backends, we may need to adapt this logic
        return
    if o.const():
        return
    if isinstance(i.name, PortViewRef):
        i = i.name.root()
    if isinstance(o.name, PortViewRef):
        o = o.name.root()
    i_defn = _get_defn(i)
    o_defn = _get_defn(o)
    if i_defn is not None and o_defn is not None and i_defn is o_defn:
        return
    i_inst = _get_inst(i)
    o_inst = _get_inst(o)
    if i_inst is not None and o_defn is not None and i_inst.defn is o_defn:
        return
    if o_inst is not None and i_defn is not None and o_inst.defn is i_defn:
        return
    if (o_inst is not None and i_inst is not None and
            o_inst.defn is i_inst.defn):
        return
    raise MagmaCompileException(
        f"Cannot wire {orig_o.debug_name} to {orig_i.debug_name} because they "
        "are not from the same definition context")



import collections
import dataclasses
import functools
import uinspect

from magma.common import Stack
from magma.config import config, RuntimeConfig


@dataclasses.dataclass
class _DebugInfo:
    filename: str
    lineno: int
    module: str


debug_info = _DebugInfo


config.register(use_uinspect=RuntimeConfig(True))


_extra_frames_to_skip_stack = Stack()


def _get_extra_frames_to_skip_stack():
    global _extra_frames_to_skip_stack
    return _extra_frames_to_skip_stack


def get_debug_info(frames_to_skip):
    extra_frames_to_skip = _get_extra_frames_to_skip_stack().peek_default(0)
    frames_to_skip += extra_frames_to_skip
    if config.use_uinspect:
        filename, lineno = uinspect.get_location(frames_to_skip)
    else:
        filename, lineno = None, None
    return debug_info(filename, lineno, None)


def debug_wire(fn):
    """
    Automatically populates the `debug_info` argument for a wire call if it's
    not already passed as an argument
    """
    def wire(i, o=None, debug_info=None):
        if debug_info is None:
            debug_info = get_debug_info(3)
        return fn(i, o, debug_info)
    return wire


def debug_unwire(fn):
    def unwire(i, o=None, debug_info=None, keep_wired_when_contexts=False):
        if debug_info is None:
            debug_info = get_debug_info(3)
        return fn(i, o, debug_info, keep_wired_when_contexts)
    return unwire


def _helper_function(fn):

    @functools.wraps(fn)
    def _wrapper(*args, **kwargs):
        extra_frames_to_skip_stack = _get_extra_frames_to_skip_stack()
        # We keep track of the previous frames_to_skip value since the stack is
        # additive.
        try:
            offset = extra_frames_to_skip_stack.peek()
        except IndexError:
            offset = 0
        # NOTE(rsetaluri): We need to skip an extra frame (1 + 1 = 2) since we
        # are returning a wrapped function (decorated).
        extra_frames_to_skip_stack.push(2 + offset)
        try:
            return fn(*args, **kwargs)
        except:
            raise
        finally:
            extra_frames_to_skip_stack.pop()

    return _wrapper


magma_helper_function = _helper_function
